# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/display/notifications/default.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 366 bytes
from .all import Notifications
from .meta import DefaultNotificationsMeta


class DefaultNotifications(Notifications, metaclass=DefaultNotificationsMeta):
    pass



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/mode.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 21350 bytes
from future.utils import iteritems

from ..base import (
    NamedTuple,
    depends,
    infinite_context_manager,
    is_contextmanager,
    is_iterable,
    lazy_attribute,
    listenable_property,
    listens,
    old_hasattr,
    task,
)
from . import defaults
from .component import Component
from .control import ButtonControl, ButtonControlBase, control_color
from .layer import CompoundLayer, Layer
from .resource import StackingResource


def tomode(thing):
    if thing is None:
        return Mode()
    if isinstance(thing, Mode):
        return thing
    if old_hasattr(thing, "set_enabled"):
        return EnablingMode(thing)
    if isinstance(thing, tuple):
        if len(thing) == 2:
            if isinstance(thing[0], Component):
                if isinstance(thing[1], (Layer, CompoundLayer)):
                    return LayerMode(*thing)
            if callable(thing[0]):
                if callable(thing[1]):
                    mode = Mode()
                    mode.enter_mode, mode.leave_mode = thing
                    return mode
    if is_iterable(thing):
        return CompoundMode(*thing)
    if is_contextmanager(thing):
        return ContextManagerMode(thing)
    if callable(thing):
        mode = Mode()
        mode.enter_mode = thing
        return mode
    return thing


def to_camel_case_name(mode_name, separator=""):
    return separator.join(s.capitalize() for s in mode_name.split("_"))


def pop_last_mode(component, mode):
    if len(component.active_modes) > 1:
        component.pop_mode(mode)


class Mode:
    def enter_mode(self):
        pass

    def leave_mode(self):
        pass

    def __enter__(self):
        self.enter_mode()

    def __exit__(self, *a):
        return self.leave_mode()


class ContextManagerMode(Mode):
    def __init__(self, context_manager=None, *a, **k):
        (super().__init__)(*a, **k)
        self._context_manager = context_manager

    def enter_mode(self):
        self._context_manager.__enter__()

    def leave_mode(self):
        self._context_manager.__exit__(None, None, None)

    def __exit__(self, exc_type, exc_value, traceback):
        return self._context_manager.__exit__(exc_type, exc_value, traceback)


def generator_mode(function):
    makecontext = infinite_context_manager(function)
    return lambda *a, **k: ContextManagerMode(makecontext(*a, **k))


class EnablingMode(Mode):
    def __init__(self, enableable=None, *a, **k):
        (super().__init__)(*a, **k)
        self._enableable = enableable

    def enter_mode(self):
        self._enableable.set_enabled(True)

    def leave_mode(self):
        self._enableable.set_enabled(False)


class LazyEnablingMode(Mode):
    def __init__(self, factory=None, *a, **k):
        (super().__init__)(*a, **k)
        self._factory = factory

    @lazy_attribute
    def enableable(self):
        return self._factory()

    def enter_mode(self):
        self.enableable.set_enabled(True)

    def leave_mode(self):
        self.enableable.set_enabled(False)


class LayerModeBase(Mode):
    def __init__(self, component=None, layer=None, *a, **k):
        (super().__init__)(*a, **k)
        self._component = component
        self._layer = layer

    def _get_component(self):
        if callable(self._component):
            return self._component()
        return self._component


class LayerMode(LayerModeBase):
    def enter_mode(self):
        self._get_component().layer = self._layer

    def leave_mode(self):
        self._get_component().layer = None


class AddLayerMode(LayerModeBase):
    def enter_mode(self):
        self._layer.grab(self._get_component())

    def leave_mode(self):
        self._layer.release(self._get_component())


class CompoundMode(Mode):
    def __init__(self, *modes, **k):
        (super().__init__)(**k)
        self._modes = list(map(tomode, modes))

    def enter_mode(self):
        for mode in self._modes:
            mode.enter_mode()

    def leave_mode(self):
        for mode in reversed(self._modes):
            mode.leave_mode()


class SetAttributeMode(Mode):
    def __init__(self, obj=None, attribute=None, value=None, *a, **k):
        (super().__init__)(*a, **k)
        self._obj = obj
        self._attribute = attribute
        self._old_value = None
        self._value = value

    def _get_object(self):
        if callable(self._obj):
            return self._obj()
        return self._obj

    def enter_mode(self):
        self._old_value = getattr(self._get_object(), self._attribute, None)
        setattr(self._get_object(), self._attribute, self._value)

    def leave_mode(self):
        if getattr(self._get_object(), self._attribute) == self._value:
            setattr(self._get_object(), self._attribute, self._old_value)


class DelayMode(Mode):
    @depends(parent_task_group=None)
    def __init__(self, mode=None, delay=None, parent_task_group=None, *a, **k):
        (super().__init__)(*a, **k)
        delay = delay if delay is not None else defaults.MOMENTARY_DELAY
        self._mode = tomode(mode)
        self._mode_entered = False
        self._delay_task = parent_task_group.add(task.sequence(task.wait(delay), task.run(self._enter_mode_delayed)))
        self._delay_task.kill()

    def _enter_mode_delayed(self):
        self._mode_entered = True
        self._mode.enter_mode()

    def enter_mode(self):
        self._delay_task.restart()

    def leave_mode(self):
        if self._mode_entered:
            self._mode.leave_mode()
            self._mode_entered = False
        self._delay_task.kill()


class ModeButtonControl(ButtonControlBase):
    class State(ButtonControlBase.State):
        mode_selected_color = control_color("DefaultButton.On")
        mode_unselected_color = control_color("DefaultButton.Off")
        mode_group_active_color = control_color("DefaultButton.On")

        def __init__(
            self,
            modes_component=None,
            mode_name=None,
            mode_selected_color=None,
            mode_unselected_color=None,
            mode_group_active_color=None,
            *a,
            **k,
        ):
            self._modes_component = modes_component
            self._mode_name = mode_name
            (super(ModeButtonControl.State, self).__init__)(*a, **k)
            if mode_selected_color is not None:
                self.mode_selected_color = mode_selected_color
            if mode_unselected_color is not None:
                self.mode_unselected_color = mode_unselected_color
            if mode_group_active_color is not None:
                self.mode_group_active_color = mode_group_active_color
            self._State__on_selected_mode_changed.subject = self._modes_component

        @property
        def mode_name(self):
            return self._mode_name

        @listens("selected_mode")
        def __on_selected_mode_changed(self, mode):
            self._send_current_color()

        def _send_button_color(self):
            selected_mode = self._modes_component.selected_mode
            groups = self._modes_component.get_mode_groups(self._mode_name)
            selected_groups = self._modes_component.get_mode_groups(selected_mode)
            if selected_mode == self._mode_name:
                self._control_element.set_light(self.mode_selected_color)
            else:
                if bool(groups & selected_groups):
                    self._control_element.set_light(self.mode_group_active_color)
                else:
                    self._control_element.set_light(self.mode_unselected_color)


class ModeButtonBehaviour:
    def press_immediate(self, component, mode):
        pass

    def release_immediate(self, component, mode):
        pass

    def press_delayed(self, component, mode):
        pass

    def release_delayed(self, component, mode):
        pass

    def update_button(self, component, mode, selected_mode):
        pass


class ImmediateBehaviour(ModeButtonBehaviour):
    def press_immediate(self, component, mode):
        component.push_mode(mode)


class LatchingBehaviour(ImmediateBehaviour):
    def release_immediate(self, component, mode):
        component.pop_unselected_modes()

    def release_delayed(self, component, mode):
        pop_last_mode(component, mode)


class MomentaryBehaviour(ImmediateBehaviour):
    def release_immediate(self, component, mode):
        pop_last_mode(component, mode)

    def release_delayed(self, component, mode):
        pop_last_mode(component, mode)


class ReenterBehaviour(LatchingBehaviour):
    def __init__(self, on_reenter=None, *a, **k):
        (super().__init__)(*a, **k)
        if on_reenter is not None:
            self.on_reenter = on_reenter

    def press_immediate(self, component, mode):
        was_active = component.selected_mode == mode
        super().press_immediate(component, mode)
        if was_active:
            self.on_reenter()

    def on_reenter(self):
        pass


def make_mode_button_control(modes_component, mode_name, behaviour, **k):
    button_control = ModeButtonControl(modes_component=modes_component, mode_name=mode_name, **k)

    @button_control.pressed
    def button_control(modes_component, button):
        behaviour.press_immediate(modes_component, mode_name)

    @button_control.pressed_delayed
    def button_control(modes_component, button):
        behaviour.press_delayed(modes_component, mode_name)

    @button_control.released_immediately
    def button_control(modes_component, button):
        behaviour.release_immediate(modes_component, mode_name)

    @button_control.released_delayed
    def button_control(modes_component, button):
        behaviour.release_delayed(modes_component, mode_name)

    return button_control


class _ModeEntry(NamedTuple):
    mode = None
    groups = set()
    cycle_mode_button_color = None
    listens = None


class NullModes:
    selected_mode = None


class ModesComponent(Component):
    cycle_mode_button = ButtonControl()
    default_behaviour = LatchingBehaviour()

    def __init__(self, enable_skinning=False, support_momentary_mode_cycling=True, *a, **k):
        (super().__init__)(*a, **k)
        self._enable_skinning = enable_skinning
        self._support_momentary_mode_cycling = support_momentary_mode_cycling
        self._last_toggle_value = 0
        self._mode_toggle = None
        self._mode_list = []
        self._mode_map = {}
        self._last_selected_mode = None
        self._mode_stack = StackingResource(self._do_enter_mode, self._do_leave_mode)

    def disconnect(self):
        self._mode_stack.release_all()
        super().disconnect()

    @listenable_property
    def selected_mode(self):
        return self._mode_stack.owner or self._last_selected_mode

    @selected_mode.setter
    def selected_mode(self, mode):
        if self.is_enabled():
            if self.selected_mode != mode:
                if mode is not None:
                    self.push_mode(mode)
                    self.pop_unselected_modes()
                else:
                    self._mode_stack.release_all()
        else:
            self._last_selected_mode = mode

    @property
    def selected_groups(self):
        entry = self._mode_map.get(self.selected_mode, None)
        if entry:
            return entry.groups
        return set()

    @property
    def active_modes(self):
        return self._mode_stack.clients

    def push_mode(self, mode):
        self._mode_stack.grab(mode)

    def pop_mode(self, mode):
        self._mode_stack.release(mode)

    def pop_groups(self, groups):
        if not isinstance(groups, set):
            groups = set(groups)
        for client in self._mode_stack.clients:
            if self.get_mode_groups(client) & groups:
                self._mode_stack.release(client)

    def pop_unselected_modes(self):
        self._mode_stack.release_stacked()

    def add_mode(self, name, mode_or_component, cycle_mode_button_color=None, groups=None, behaviour=None):
        if groups is None:
            groups = set()
        if not isinstance(groups, set):
            groups = set(groups)
        mode = tomode(mode_or_component)
        behaviour = behaviour if behaviour is not None else self.default_behaviour
        self._mode_list.append(name)
        self._mode_map[name] = _ModeEntry(
            mode=mode,
            cycle_mode_button_color=cycle_mode_button_color,
            behaviour=behaviour,
            groups=groups,
        )
        self.add_mode_button_control(name, behaviour)

    @property
    def modes(self):
        return self._mode_list

    def get_mode_groups(self, name):
        entry = self._mode_map.get(name, None)
        if entry:
            return entry.groups
        return set()

    def add_mode_button_control(self, mode_name, behaviour):
        colors = {}
        if self._enable_skinning:
            mode_color_basename = "Mode." + to_camel_case_name(mode_name)
            colors = {
                "mode_selected_color": mode_color_basename + ".On",
                "mode_unselected_color": mode_color_basename + ".Off",
                "mode_group_active_color": mode_color_basename + ".On",
            }
        button_control = make_mode_button_control(self, mode_name, behaviour, **colors)
        self.add_control(f"{mode_name}_button", button_control)
        self._update_mode_buttons(self.selected_mode)

    def _get_mode_behaviour(self, name):
        entry = self._mode_map.get(name, None)
        if entry is not None:
            return entry.behaviour
        return self.default_behaviour

    def get_mode(self, name):
        entry = self._mode_map.get(name, None)
        return entry and entry.mode

    def get_mode_button(self, name):
        return getattr(self, f"{name}_button")

    def _update_mode_buttons(self, selected):
        if self.is_enabled():
            for name, entry in iteritems(self._mode_map):
                self._get_mode_behaviour(name).update_button(self, name, selected)

    @cycle_mode_button.pressed
    def cycle_mode_button(self, button):
        if len(self._mode_list):
            self.cycle_mode(1)

    @cycle_mode_button.released_delayed
    def cycle_mode_button(self, button):
        if self._support_momentary_mode_cycling:
            if len(self._mode_list):
                if self.selected_mode != self._mode_list[0]:
                    self.cycle_mode(-1)

    def _update_cycle_mode_button(self, selected):
        entry = self._mode_map.get(selected)
        color = entry.cycle_mode_button_color if entry else None
        if color is not None:
            self.cycle_mode_button.color = color

    def cycle_mode(self, delta=1):
        current_index = self._mode_list.index(self.selected_mode) if self.selected_mode else -delta
        current_index = (current_index + delta) % len(self._mode_list)
        self.selected_mode = self._mode_list[current_index]

    def _do_enter_mode(self, name):
        entry = self._mode_map[name]
        entry.mode.enter_mode()
        self._update_mode_buttons(name)
        self._update_cycle_mode_button(name)
        self.notify_selected_mode(name)

    def _do_leave_mode(self, name):
        self._mode_map[name].mode.leave_mode()
        if self._mode_stack.stack_size == 0:
            self._update_mode_buttons(None)
            self._update_cycle_mode_button(None)
            self.notify_selected_mode(None)

    def on_enabled_changed(self):
        super().on_enabled_changed()
        if not self.is_enabled():
            self._last_selected_mode = self.selected_mode
            self._mode_stack.release_all()
        else:
            if self._last_selected_mode:
                self.push_mode(self._last_selected_mode)


class EnablingModesComponent(ModesComponent):
    def __init__(self, component=None, enabled_color="DefaultButton.On", disabled_color="DefaultButton.Off", *a, **k):
        (super().__init__)(*a, **k)
        component.set_enabled(False)
        self.add_mode("disabled", None, disabled_color)
        self.add_mode("enabled", component, enabled_color)
        self.selected_mode = "disabled"



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/components/clip_actions.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 5309 bytes
from Live.Song import RecordingQuantization

from ...base import depends, listens
from ...live import action, display_name, liveobj_valid
from .. import Component
from ..controls import ButtonControl
from ..display import Renderable

QUANTIZATION_OPTION_NAMES = {
    RecordingQuantization.rec_q_no_q: "None",
    RecordingQuantization.rec_q_quarter: "1/4",
    RecordingQuantization.rec_q_eight: "1/8",
    RecordingQuantization.rec_q_eight_triplet: "1/8T",
    RecordingQuantization.rec_q_eight_eight_triplet: "1/8+T",
    RecordingQuantization.rec_q_sixtenth: "1/16",
    RecordingQuantization.rec_q_sixtenth_triplet: "1/16T",
    RecordingQuantization.rec_q_sixtenth_sixtenth_triplet: "1/16+T",
    RecordingQuantization.rec_q_thirtysecond: "1/32",
}


class ClipActionsComponent(Component, Renderable):
    delete_button = ButtonControl(color="ClipActions.Delete", pressed_color="ClipActions.DeletePressed")
    double_button = ButtonControl(color="ClipActions.Double", pressed_color="ClipActions.DoublePressed")
    duplicate_button = ButtonControl(color="ClipActions.Duplicate", pressed_color="ClipActions.DuplicatePressed")
    quantize_button = ButtonControl(color="ClipActions.Quantize", pressed_color="ClipActions.QuantizePressed")

    @depends(target_track=None)
    def __init__(self, target_track=None, *a, **k):
        (super().__init__)(a, name="Clip_Actions", **k)
        self._target_track = target_track
        self._ClipActionsComponent__on_target_clip_changed.subject = target_track
        self._ClipActionsComponent__on_target_clip_recording_changed.subject = target_track
        self._ClipActionsComponent__on_target_clip_playing_status_changed.subject = target_track
        self._update_action_buttons()
        self._quantization_value = RecordingQuantization.rec_q_sixtenth
        self._ClipActionsComponent__on_record_quantization_changed.subject = self.song
        self._ClipActionsComponent__on_record_quantization_changed()

    @delete_button.pressed
    def delete_button(self, _):
        self.notify(self.notifications.Clip.delete, display_name(self._target_track.target_clip))
        action.delete(self._target_track.target_clip)

    @double_button.pressed
    def double_button(self, _):
        action.duplicate_loop(self._target_track.target_clip)
        self.notify(self.notifications.Clip.double_loop)

    @duplicate_button.pressed
    def duplicate_button(self, _):
        if action.duplicate_clip_special(self._target_track.target_clip):
            self.notify(self.notifications.Clip.duplicate, display_name(self._target_track.target_clip))

    @quantize_button.pressed
    def quantize_button(self, _):
        self._quantize_clip(self._target_track.target_clip)

    def _quantize_clip(self, clip):
        clip.quantize(self._quantization_value, 1.0)
        self.notify(
            self.notifications.Clip.quantize,
            display_name(clip),
            QUANTIZATION_OPTION_NAMES[self._quantization_value],
        )

    def _update_action_buttons(self):
        self._update_delete_button()
        self._update_double_button()
        self._update_duplicate_button()
        self._update_quantize_button()

    def _update_delete_button(self):
        self.delete_button.enabled = self._get_target_clip() is not None

    def _update_double_button(self):
        clip = self._get_target_clip()
        self.double_button.enabled = clip is not None and clip.is_midi_clip

    def _update_duplicate_button(self):
        self.duplicate_button.enabled = self._get_target_clip() is not None

    def _update_quantize_button(self):
        self.quantize_button.enabled = self._get_target_clip() is not None

    def _get_target_clip(self):
        clip = self._target_track.target_clip
        if liveobj_valid(clip):
            if not clip.is_recording:
                if not clip.will_record_on_start:
                    return clip
                return None
            return None
        return None

    @listens("target_clip")
    def __on_target_clip_changed(self):
        self._update_action_buttons()

    @listens("target_clip.is_recording")
    def __on_target_clip_recording_changed(self):
        self._update_action_buttons()

    @listens("target_clip.playing_status")
    def __on_target_clip_playing_status_changed(self):
        self._update_action_buttons()

    @listens("midi_recording_quantization")
    def __on_record_quantization_changed(self):
        quantization_value = self.song.midi_recording_quantization
        if quantization_value != RecordingQuantization.rec_q_no_q:
            self._quantization_value = quantization_value



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/identifiable_control_surface.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 2886 bytes
import logging

from ..base import task
from . import midi
from .control_surface import ControlSurface

logger = logging.getLogger(__name__)


class IdentifiableControlSurface(ControlSurface):
    identity_request_delay = 0.0
    identity_request = midi.SYSEX_IDENTITY_REQUEST_MESSAGE

    def __init__(self, product_id_bytes=None, *a, **k):
        (super().__init__)(*a, **k)
        self._product_id_bytes = product_id_bytes
        self._identity_response_pending = False
        self._request_task = self._tasks.add(
            task.sequence(task.wait(self.identity_request_delay), task.run(self._send_identity_request)),
        )
        self._request_task.kill()

    def on_identified(self, response_bytes):
        self.refresh_state()

    def port_settings_changed(self):
        self._request_task.restart()

    def process_midi_bytes(self, midi_bytes, midi_processor):
        if midi.is_sysex(midi_bytes) and self._is_identity_response(midi_bytes):
            product_id_bytes = self._extract_product_id_bytes(midi_bytes)
            if product_id_bytes == self._product_id_bytes:
                self._request_task.kill()
                if self._identity_response_pending:
                    self.on_identified(midi_bytes)
                    self._identity_response_pending = False
            else:
                logger.error(
                    "MIDI device responded with wrong product id (%s != %s).",
                    str(self._product_id_bytes),
                    str(product_id_bytes),
                )
        else:
            super().process_midi_bytes(midi_bytes, midi_processor)

    def _is_identity_response(self, midi_bytes):
        return midi_bytes[3:5] == (midi.SYSEX_GENERAL_INFO, midi.SYSEX_IDENTITY_RESPONSE_ID)

    def _extract_product_id_bytes(self, midi_bytes):
        return midi_bytes[5 : 5 + len(self._product_id_bytes)]

    def _send_identity_request(self):
        self._identity_response_pending = True
        self._send_midi(self.identity_request)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/components/device.py
# Compiled at: 2023-12-18 13:26:15
# Size of source mod 2**32: 12300 bytes
from typing import cast

from ableton.v2.control_surface.simpler_slice_nudging import SimplerSliceNudging

from ...base import depends, find_if, listenable_property, listens
from ...live import deduplicate_parameters, liveobj_changed, liveobj_valid
from .. import DEFAULT_BANK_SIZE, BankingInfo, Component, ParameterProvider, create_parameter_bank
from ..controls import MappedButtonControl, ToggleButtonControl
from ..default_bank_definitions import BANK_DEFINITIONS
from ..display import Renderable
from ..parameter_info import ParameterInfo
from ..parameter_mapping_sensitivities import (
    DEFAULT_CONTINUOUS_PARAMETER_SENSITIVITY,
    DEFAULT_QUANTIZED_PARAMETER_SENSITIVITY,
    parameter_mapping_sensitivities,
)
from .device_bank_navigation import DeviceBankNavigationComponent
from .device_parameters import DeviceParametersComponent


def get_on_off_parameter(device):
    if liveobj_valid(device):
        return find_if(
            lambda p: p.original_name.startswith("Device On") and liveobj_valid(p) and p.is_enabled,
            device.parameters,
        )
    return None


def create_device_decorator_factory():
    pass


class DeviceComponent(ParameterProvider, Component, Renderable):
    device_on_off_button = MappedButtonControl(color="Device.Off", on_color="Device.On")
    device_lock_button = ToggleButtonControl(color="Device.LockOff", on_color="Device.LockOn")

    @depends(device_provider=None, device_bank_registry=None, toggle_lock=None, show_message=None)
    def __init__(
        self,
        name="Device",
        continuous_parameter_sensitivity=DEFAULT_CONTINUOUS_PARAMETER_SENSITIVITY,
        quantized_parameter_sensitivity=DEFAULT_QUANTIZED_PARAMETER_SENSITIVITY,
        parameters_component_type=None,
        bank_size=DEFAULT_BANK_SIZE,
        bank_definitions=None,
        bank_navigation_component_type=None,
        device_provider=None,
        device_bank_registry=None,
        device_decorator_factory=None,
        toggle_lock=None,
        show_message=None,
        *a,
        **k,
    ):
        self._decorated_device = None
        self._provided_parameters = []
        self._device_provider = device_provider
        self._device_bank_registry = device_bank_registry
        self._decorator_factory = create_device_decorator_factory(device_decorator_factory, bank_definitions)
        self._parameter_mapping_sensitivities = parameter_mapping_sensitivities(
            continuous_parameter_sensitivity=continuous_parameter_sensitivity,
            quantized_parameter_sensitivity=quantized_parameter_sensitivity,
        )
        parameters_component_type = parameters_component_type or DeviceParametersComponent
        self._parameters_component = parameters_component_type()
        self._parameters_component.parameter_provider = self
        self._bank = None
        self._banking_info = BankingInfo((bank_definitions or BANK_DEFINITIONS), bank_size=bank_size)
        bank_navigation_component_type = bank_navigation_component_type or DeviceBankNavigationComponent
        self._bank_navigation_component = bank_navigation_component_type(
            banking_info=(self._banking_info),
            device_bank_registry=device_bank_registry,
        )
        (super().__init__)(a, name=name, **k)
        self._toggle_lock = toggle_lock
        self._show_message = show_message
        self._slice_nudging = self.register_disconnectable(SimplerSliceNudging())
        self.add_children(self._parameters_component, self._bank_navigation_component)
        self._DeviceComponent__on_provided_device_changed.subject = device_provider
        self._DeviceComponent__on_provided_device_changed()
        self.register_slot(self._device_provider, self._update_device_lock_button, "is_locked_to_device")
        self._update_device_lock_button()

    @property
    def parameters(self):
        return self._provided_parameters

    @listenable_property
    def device(self):
        return self._decorated_device

    @device.setter
    def device(self, device):
        self._device_provider.device = device

    @listenable_property
    def bank_name(self):
        if self.device:
            return self._current_bank_details()[0]
        return ""

    def set_parameter_controls(self, controls):
        self._parameters_component.set_parameter_controls(controls)
        self._show_device_and_bank_info()

    def __getattr__(self, name):
        if name.startswith("set_"):
            if "bank" in name:
                return getattr(self._bank_navigation_component, name)
        raise AttributeError

    @device_lock_button.toggled
    def device_lock_button(self, *_):
        self._toggle_lock()
        if liveobj_valid(self.device):
            self.notify(
                self.notifications.Device.lock,
                cast(str, self.device.name),
                self._device_provider.is_locked_to_device,
            )

    def _create_parameter_info(self, parameter, name):
        default, fine_grain = self._parameter_mapping_sensitivities(parameter, self.device)
        return ParameterInfo(
            parameter=parameter,
            name=name,
            default_encoder_sensitivity=default,
            fine_grain_encoder_sensitivity=fine_grain,
        )

    def _set_device(self, device):
        self._set_decorated_device(self._get_decorated_device(device))
        device_bank_registry = self._device_bank_registry
        self._DeviceComponent__on_bank_changed.subject = device_bank_registry
        self._set_bank_index(device_bank_registry.get_device_bank(device))
        self._update_parameters()
        self._DeviceComponent__on_parameters_changed_in_device.subject = device
        self.device_on_off_button.mapped_parameter = get_on_off_parameter(device)
        self.notify_device()
        self.notify_bank_name()

    def _get_decorated_device(self, device):
        if self._decorator_factory is not None:
            return self._decorator_factory.decorate(device)
        return device

    def _set_decorated_device(self, decorated_device):
        self._decorated_device = decorated_device
        self._slice_nudging.set_device(decorated_device)
        self._setup_bank(decorated_device)
        self._DeviceComponent__on_bank_parameters_changed.subject = self._bank

    def _on_device_changed(self, device):
        current_device = getattr(self.device, "_live_object", self.device)
        if liveobj_changed(current_device, device):
            self._set_device(device)

    def _setup_bank(self, device, bank_factory=create_parameter_bank):
        if self._bank is not None:
            self.disconnect_disconnectable(self._bank)
            self._bank = None
        if liveobj_valid(device):
            self._bank = self.register_disconnectable(bank_factory(device, self._banking_info))
        self._bank_navigation_component.bank_provider = self._bank

    def _set_bank_index(self, bank):
        if self._bank is not None:
            self._bank.index = bank
        if liveobj_valid(self.device):
            self.notify_bank_name()
            self._show_device_and_bank_info()

    def _current_bank_details(self):
        if self._bank is not None:
            return (self._bank.name, self._bank.parameters)
        return ("", [None] * self._parameters_component.controls.control_count)

    def _show_device_and_bank_info(self):
        device = self.device
        if liveobj_valid(device):
            if self._parameters_component.controls[0].control_element:
                self._show_message(f"Controlling {device.name}: {self.bank_name}")

    def _get_provided_parameters(self):
        _, parameters = self._current_bank_details() if self.device else (None, ())
        return deduplicate_parameters([self._create_parameter_info(param, name) for param, name in parameters])

    def _update_parameters(self):
        self._provided_parameters = self._get_provided_parameters()
        self.notify_parameters()

    def _update_device_lock_button(self):
        self.device_lock_button.is_on = self._device_provider.is_locked_to_device

    @listens("device")
    def __on_provided_device_changed(self):
        self._on_device_changed(self._device_provider.device)

    @listens("device_bank")
    def __on_bank_changed(self, device, bank):
        if device == self.device:
            self._set_bank_index(bank)

    @listens("parameters")
    def __on_parameters_changed_in_device(self):
        self._update_parameters()

    @listens("parameters")
    def __on_bank_parameters_changed(self):
        self._update_parameters()
        self._device_bank_registry.set_device_bank(self.device, self._bank.index)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/components/scene.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 5236 bytes
from ...base import depends, listens
from ...live import action, display_name, liveobj_changed, liveobj_valid, scene_index
from .. import Component
from ..controls import ButtonControl
from ..display import Renderable
from ..skin import LiveObjSkinEntry
from . import ClipSlotComponent


class SceneComponent(Component, Renderable):
    launch_button = ButtonControl()
    select_button = ButtonControl(color=None)
    delete_button = ButtonControl(color=None)
    duplicate_button = ButtonControl(color=None)
    include_in_top_level_state = False

    @depends(session_ring=None)
    def __init__(self, session_ring=None, clip_slot_component_type=None, *a, **k):
        (super().__init__)(*a, **k)
        self._session_ring = session_ring
        self._scene = None
        clip_slot_component_type = clip_slot_component_type or ClipSlotComponent
        self._clip_slots = [clip_slot_component_type(parent=self) for _ in range(session_ring.num_tracks)]
        self.register_slot(session_ring, self._reassign_clip_slots, "tracks")

    @property
    def scene(self):
        return self._scene

    def set_scene(self, scene):
        if liveobj_changed(scene, self._scene):
            self._scene = scene
            self._SceneComponent__on_is_triggered_changed.subject = scene
            self._SceneComponent__on_scene_color_changed.subject = scene
            self.update()

    def clip_slot(self, index):
        return self._clip_slots[index]

    def set_launch_button(self, button):
        self.launch_button.set_control_element(button)
        self.update()

    @launch_button.pressed
    def launch_button(self, _):
        self._on_launch_button_pressed()

    def _on_launch_button_pressed(self):
        scene_name = display_name(self._scene) if liveobj_valid(self._scene) else ""
        if self.select_button.is_pressed:
            if action.select(self._scene):
                self.notify(self.notifications.Scene.select, scene_name)
        else:
            if self.duplicate_button.is_pressed:
                action.duplicate(self._scene)
            else:
                if self.delete_button.is_pressed:
                    if action.delete(self._scene):
                        self.notify(self.notifications.Scene.delete, scene_name)
                else:
                    self._do_launch_scene()

    def _do_launch_scene(self):
        action.fire((self._scene), button_state=True)

    @launch_button.released
    def launch_button(self, _):
        self._on_launch_button_released()

    def _on_launch_button_released(self):
        if self.launch_button.is_momentary:
            if not self._any_modifier_pressed():
                action.fire((self._scene), button_state=False)

    def _any_modifier_pressed(self):
        return self.select_button.is_pressed or self.delete_button.is_pressed or self.duplicate_button.is_pressed

    def update(self):
        super().update()
        self._reassign_clip_slots()
        self._update_launch_button_color()

    def _update_launch_button_color(self):
        value_to_send = "Session.NoScene"
        if liveobj_valid(self._scene):
            value_to_send = self._feedback_value()
        self.launch_button.color = value_to_send

    def _feedback_value(self):
        value = "Session.Scene"
        if self._scene.is_triggered:
            value = "Session.SceneTriggered"
        return LiveObjSkinEntry(value, self._scene)

    def _reassign_clip_slots(self):
        if liveobj_valid(self._scene) and self.is_enabled():
            scene_offset = scene_index(self._scene)
            regular_tracks = self.song.tracks
            for slot_wrapper, track in zip(self._clip_slots, self._session_ring.tracks):
                if track in regular_tracks:
                    slot_wrapper.set_clip_slot(track.clip_slots[scene_offset])
                else:
                    slot_wrapper.set_non_player_track(track)

        else:
            for slot in self._clip_slots:
                slot.set_clip_slot(None)

    @listens("is_triggered")
    def __on_is_triggered_changed(self):
        self._update_launch_button_color()

    @listens("color")
    def __on_scene_color_changed(self):
        self._update_launch_button_color()



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/base/abl_signal.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 3065 bytes
from functools import partial

from .util import find_if, nop


def default_combiner(results):
    for _ in results:
        pass


class Slot:
    def __init__(self, callback=None, *a, **k):
        (super().__init__)(*a, **k)
        self.callback = callback

    def __call__(self, *a, **k):
        return (self.callback)(*a, **k)

    def __eq__(self, other):
        return id(self) == id(other) or self.callback == other

    def __hash__(self):
        return hash((self.callback,))


class IdentifyingSlot(Slot):
    def __init__(self, sender=None, *a, **k):
        (super().__init__)(*a, **k)
        self.sender = sender

    def __call__(self, *a, **k):
        (self.callback)(*(*a, self.sender), **k)


class Signal:
    def __init__(self, combiner=default_combiner, sender=None, *a, **k):
        (super().__init__)(*a, **k)
        self._slots = []
        self._combiner = combiner

    def connect(self, slot, in_front=False, sender=None):
        if slot not in self._slots:
            slot = IdentifyingSlot(sender, slot) if sender is not None else Slot(slot)
            if in_front:
                self._slots.insert(0, slot)
            else:
                self._slots.append(slot)
        else:
            slot = find_if(lambda x: x == slot, self._slots)
        return slot

    def disconnect(self, slot):
        if slot in self._slots:
            self._slots.remove(slot)

    def disconnect_all(self):
        self._slots = []

    @property
    def count(self):
        return len(self._slots)

    def is_connected(self, slot):
        return slot in self._slots

    def __call__(self, *a, **k):
        return self._combiner(_slot_notification_generator(self._slots, a, k))


def _slot_notification_generator(slots, args, kws):
    for slot in slots:
        yield slot(*args, **kws)


def short_circuit_combiner(slot_results):
    return find_if(nop, slot_results)


short_circuit_signal = partial(Signal, short_circuit_combiner)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/components/scroll.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 5388 bytes
from ...base import task
from .. import MOMENTARY_DELAY, Component
from ..controls import ButtonControl, StepEncoderControl


class Scrollable:
    can_scroll_up = NotImplemented
    can_scroll_down = NotImplemented
    scroll_up = NotImplemented
    scroll_down = NotImplemented


class ScrollComponent(Component, Scrollable):
    scrolling_delay = MOMENTARY_DELAY
    scrolling_step_delay = 0.1
    scroll_encoder = StepEncoderControl(num_steps=64)
    scroll_up_button = ButtonControl()
    scroll_down_button = ButtonControl()

    def __init__(self, scrollable=None, scroll_skin_name=None, *a, **k):
        (super().__init__)(*a, **k)
        if scroll_skin_name:
            pressed_colored = f"{scroll_skin_name}Pressed"
            self.scroll_up_button.color = scroll_skin_name
            self.scroll_down_button.color = scroll_skin_name
            self.scroll_up_button.pressed_color = pressed_colored
            self.scroll_down_button.pressed_color = pressed_colored
        self._scroll_task_up = self._make_scroll_task(self._do_scroll_up)
        self._scroll_task_down = self._make_scroll_task(self._do_scroll_down)
        self._scrollable = scrollable or self

    def _make_scroll_task(self, scroll_step):
        t = self._tasks.add(
            task.sequence(
                task.wait(self.scrolling_delay),
                task.loop(task.wait(self.scrolling_step_delay), task.run(scroll_step)),
            ),
        )
        t.kill()
        return t

    def can_scroll_up(self):
        return self._scrollable.can_scroll_up()

    def can_scroll_down(self):
        return self._scrollable.can_scroll_down()

    def scroll_up(self):
        return self._scrollable.scroll_up()

    def scroll_down(self):
        return self._scrollable.scroll_down()

    def set_scroll_up_button(self, button):
        self.scroll_up_button.set_control_element(button)
        self._update_scroll_buttons()

    def set_scroll_down_button(self, button):
        self.scroll_down_button.set_control_element(button)
        self._update_scroll_buttons()

    def set_scroll_encoder(self, encoder):
        self.scroll_encoder.set_control_element(encoder)

    def _update_scroll_buttons(self):
        self.scroll_up_button.enabled = self.can_scroll_up()
        self.scroll_down_button.enabled = self.can_scroll_down()

    @scroll_up_button.pressed
    def scroll_up_button(self, button):
        self._on_scroll_pressed(button, self._do_scroll_up, self._scroll_task_up)

    @scroll_up_button.released
    def scroll_up_button(self, _):
        self._on_scroll_released(self._scroll_task_up)

    @scroll_down_button.pressed
    def scroll_down_button(self, button):
        self._on_scroll_pressed(button, self._do_scroll_down, self._scroll_task_down)

    @scroll_down_button.released
    def scroll_down_button(self, _):
        self._on_scroll_released(self._scroll_task_down)

    @scroll_encoder.value
    def scroll_encoder(self, value, _):
        if value < 0:
            if self.can_scroll_up():
                self.scroll_up()
        else:
            if self.can_scroll_down():
                self.scroll_down()

    def _do_scroll_up(self):
        self.scroll_up()
        self._update_scroll_buttons()

    def _do_scroll_down(self):
        self.scroll_down()
        self._update_scroll_buttons()

    def update(self):
        super().update()
        self._update_scroll_buttons()

    def _on_scroll_pressed(self, button, scroll_step, scroll_task):
        is_scrolling = not self._scroll_task_up.is_killed or not self._scroll_task_down.is_killed
        if not is_scrolling:
            scroll_step()
        if button.enabled:
            scroll_task.restart()
        self._ensure_scroll_one_direction()

    def _on_scroll_released(self, scroll_task):
        scroll_task.kill()
        self._ensure_scroll_one_direction()

    def _ensure_scroll_one_direction(self):
        if self.scroll_up_button.is_pressed and self.scroll_down_button.is_pressed:
            self._scroll_task_up.pause()
            self._scroll_task_down.pause()
        else:
            self._scroll_task_up.resume()
            self._scroll_task_down.resume()



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/base/util.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 23724 bytes
import sys
from contextlib import contextmanager, suppress
from functools import partial, reduce, wraps
from itertools import chain
from numbers import Number

from future.builtins import map, range
from future.moves.itertools import zip_longest
from future.utils import iteritems

PY2 = sys.version_info[0] < 3
PY3 = sys.version_info[0] >= 3


def clamp(val, minv, maxv):
    return max(minv, min(val, maxv))


def linear(minv, maxv, val):
    return minv + (maxv - minv) * val


def nop(*a, **k):
    if a:
        return a[0]
    return None


def negate(value):
    return not value


def old_round(value):
    if round(value + 1) - round(value) != 1:
        return value + abs(value) / value * 0.5
    return round(value)


def const(value):
    return lambda *a, **k: value


def in_range(value, lower_bound, upper_open_bound):
    if not isinstance(value, Number):
        return False
    return value >= lower_bound and value < upper_open_bound


def sign(value):
    if value >= 0.0:
        return 1.0
    return -1.0


def to_slice(obj):
    if isinstance(obj, slice):
        return obj
    if obj != -1:
        return slice(obj, obj + 1)
    return slice(obj, None)


def slice_size(slice, width):
    return len(range(width)[slice])


def chunks(lst, chunk_size):
    for i in range(0, len(lst), chunk_size):
        yield lst[i : i + chunk_size]


def maybe(fn):
    return lambda x: fn(x) if x is not None else None


def memoize(function):
    memoized = {}

    @wraps(function)
    def wrapper(*args):
        try:
            ret = memoized[args]
        except KeyError:
            ret = memoized[args] = function(*args)

        return ret

    return wrapper


@memoize
def mixin(*args):
    if len(args) == 1:
        return args[0]
    name = "Mixin_{}".format("_".join(cls.__name__ for cls in args))
    return type(str(name), args, {})


def monkeypatch(target, name=None, override=False, doc=None):
    def patcher(func):
        patchname = func.__name__ if name is None else name
        if not override:
            if old_hasattr(target, patchname):
                raise TypeError(f"Class {target.__name__} already has method {patchname}")
        setattr(target, patchname, func)
        with suppress(AttributeError):
            func.__name__ = str(patchname)

        if doc is not None:
            func.__doc__ = doc
        return func

    return patcher


def monkeypatch_extend(target, name=None):
    def patcher(func):
        newfunc = func
        patchname = func.__name__ if name is None else name
        if old_hasattr(target, patchname):
            oldfunc = getattr(target, patchname)
            if not callable(oldfunc):
                raise TypeError("Can not extend non callable attribute")

            @wraps(oldfunc)
            def extended(*a, **k):
                ret = oldfunc(*a, **k)
                func(*a, **k)
                return ret

            newfunc = extended
        else:
            pass
        setattr(target, patchname, newfunc)
        return func

    return patcher


def instance_decorator(decorator):
    class Decorator:
        def __init__(self, func=nop, *args, **kws):
            self.__name__ = func.__name__
            self.__doc__ = func.__doc__
            self._data_name = "%s_%d_decorated_instance" % (func.__name__, id(self))
            self._func = func
            self._args = args
            self._kws = kws

        def __get__(self, obj, cls=None):
            if obj is None:
                return None
            data_name = self._data_name
            try:
                return obj.__dict__[data_name]
            except KeyError:
                decorated = decorator(obj, self._func, *(self._args), **self._kws)
                obj.__dict__[data_name] = decorated
                return decorated

    return Decorator


def forward_property(member):
    class Descriptor:
        def __init__(self, func_or_name):
            self._property_name = func_or_name.__name__ if callable(func_or_name) else func_or_name

        def __get__(self, obj, cls=None):
            return getattr(getattr(obj, member), self._property_name)

        def __set__(self, obj, value):
            return setattr(getattr(obj, member), self._property_name, value)

    return Descriptor


class lazy_attribute:
    def __init__(self, func, name=None):
        wraps(func)(self)
        self._func = func
        if name:
            self.__name__ = name

    def __get__(self, obj, cls=None):
        if obj is None:
            return self
        result = obj.__dict__[self.__name__] = self._func(obj)
        return result


def remove_if(predicate, lst):
    return [elem for elem in lst if not predicate(elem)]


def flatten(list):
    return chain(*list)


def group(lst, n):
    n = int(n)
    return list(zip_longest(*[lst[i::n] for i in range(n)]))


def find_if(predicate, seq):
    for x in seq:
        if predicate(x):
            return x
    return None


def index_if(predicate, seq):
    idx = 0
    for x in seq:
        if predicate(x):
            return idx
        else:
            idx += 1

    return idx


def union(a, b):
    a = dict(a)
    a.update(b)
    return a


def product(iter_a, iter_b):
    for a in iter_a:
        for b in iter_b:
            yield (a, b)


def next(iter):
    return iter.next()


def is_iterable(value):
    try:
        it = iter(value)
        return bool(it)
    except TypeError:
        return False


def recursive_map(fn, element, sequence_type=None):
    if sequence_type is None:
        return recursive_map(fn, element, type(element))
    if isinstance(element, sequence_type):
        return [recursive_map(fn, x, sequence_type) for x in element]
    return fn(element)


def is_matrix(iterable):
    if is_iterable(iterable):
        if len(iterable) > 0:
            return all(map(lambda x: is_iterable(x) and len(iterable[0]) == len(x) and len(x) > 0, iterable))
    return False


def first(seq):
    return seq[0]


def second(seq):
    return seq[1]


def third(seq):
    return seq[2]


def compose(*funcs):
    return lambda x: reduce(lambda x, f: f(x), funcs[::-1], x)


def is_contextmanager(value):
    return callable(getattr(value, "__enter__", None)) and callable(getattr(value, "__exit__", None))


def infinite_context_manager(generator):
    make_context_manager = contextmanager(generator)

    class InfiniteContextManager:
        def __enter__(self):
            self._delegate = make_context_manager()
            self._delegate.__enter__()

        def __exit__(self, type, err, trace):
            self._delegate.__exit__(type, err, trace)
            del self._delegate

    return InfiniteContextManager


@contextmanager
def aggregate_contexts(handlers):
    for handler in handlers:
        handler.__enter__()

    err = None
    exc_info = (None, None, None)
    try:
        yield
    except BaseException as e:
        try:
            exc_info = sys.exc_info()
            err = e
        finally:
            e = None
            del e

    for handler in reversed(handlers):
        if (handler.__exit__)(*exc_info):
            err = False
            exc_info = (None, None, None)

    if err:
        raise err


class BooleanContext:
    default_value = False

    def __init__(self, default_value=None, *a, **k):
        (super().__init__)(*a, **k)
        if default_value is not None:
            self.default_value = default_value
        self._current_value = self.default_value

    def __nonzero__(self):
        return bool(self._current_value)

    def __bool__(self):
        return self.__nonzero__()

    def __call__(self, update_value=None):
        return self.Manager(self, update_value)

    @property
    def value(self):
        return self._current_value

    class Manager:
        def __init__(self, managed=None, update_value=None, *a, **k):
            (super(BooleanContext.Manager, self).__init__)(*a, **k)
            self._managed = managed
            self._update_value = update_value if update_value is not None else not managed.default_value

        def __enter__(self):
            managed = self._managed
            self._old_value = managed._current_value
            managed._current_value = self._update_value
            return self

        def __exit__(self, *a, **k):
            self._managed._current_value = self._old_value


class OutermostOnlyContext:
    def __init__(self, *a, **k):
        self._entered_count = 0

    @contextmanager
    def __call__(self, context):
        self._entered_count += 1
        if self._entered_count == 1:
            with context as result:
                yield result
        else:
            yield
        self._entered_count -= 1


def dict_diff(left, right):
    dummy = object()
    return dict(filter(lambda kv: left.get(kv[0], dummy) != kv[1], iteritems(right)))


class NamedTuple:
    def __init__(self, *others, **k):
        super().__init__()
        for other in others:
            diff = dict_diff(self._eq_dict, other._eq_dict)
            self._eq_dict.update(diff)
            self.__dict__.update(diff)

        self.__dict__.update(k)
        if "_eq_dict" in self.__dict__:
            self._eq_dict.update(k)

    def __setattr__(self, name, value):
        raise AttributeError("Named tuple is constant")

    def __delattr__(self, name):
        raise AttributeError("Named tuple is constant")

    def __getitem__(self, name):
        return self.__dict__[name]

    @lazy_attribute
    def _eq_dict(self):
        def public(objdict):
            return dict(filter(lambda kv: not kv[0].startswith("_"), iteritems(objdict)))

        return reduce(
            lambda a, b: union(b, a),
            map(lambda c: public(c.__dict__), self.__class__.__mro__),
            public(self.__dict__),
        )

    def __eq__(self, other):
        return isinstance(other, NamedTuple) and self._eq_dict == other._eq_dict

    def __hash__(self):
        return hash(id(self))

    def __getstate__(self):
        res = dict(self.__dict__)
        with suppress(KeyError):
            del res["_eq_dict"]

        return res


class Slicer:
    def __init__(self, dimensions=1, extractor=nop, keys=(), *a, **k):
        (super().__init__)(*a, **k)
        self._keys = keys
        self._dimensions = dimensions
        self._extractor = extractor

    def __getitem__(self, key):
        new = key if isinstance(key, tuple) else (key,)
        keys = self._keys + new
        if len(keys) == self._dimensions:
            return (self._extractor)(*keys)
        return Slicer(dimensions=(self._dimensions), extractor=(self._extractor), keys=keys)

    def __call__(self):
        return self


get_slice = Slicer()


def slicer(dimensions):
    def decorator(extractor):
        @wraps(extractor)
        def make_slicer(*a, **k):
            return Slicer(dimensions=dimensions, extractor=partial(extractor, *a, **k))

        return make_slicer

    return decorator


def print_message(*messages):
    print(" ".join(list(map(str, messages))))


def old_hasattr(obj, attr):
    try:
        getattr(obj, attr)
        return True
    except Exception:
        return False


class overlaymap:
    def __init__(self, *maps):
        self._maps = maps

    def __getitem__(self, key):
        for m in self._maps:
            if key in m:
                return m[key]

        raise KeyError(key)

    def keys(self):
        res = set()
        for key in chain.from_iterable(self._maps):
            res.add(key)

        return list(res)

    def values(self):
        return [self[key] for key in self.keys()]

    def iteritems(self):
        for key in self.keys():
            yield (key, self[key])


def trace_value(value, msg="Value: "):
    print(msg, value)
    return value


class Bindable:
    _bound_instances = None

    def __get__(self, obj, cls=None):
        import weakref

        if obj is None:
            return self
        if self._bound_instances is None:
            self._bound_instances = weakref.WeakKeyDictionary()
        bound_dict = self._bound_instances.setdefault(obj, weakref.WeakKeyDictionary())
        try:
            bound = bound_dict[self]
        except KeyError:
            bound = self.bind(weakref.proxy(obj))
            bound_dict[self] = bound

        return bound

    def bind(self, bind_to_object):
        raise NotImplementedError



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/components/device_parameters.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 5812 bytes
from itertools import chain, repeat

from future.moves.itertools import zip_longest
from past.builtins import unicode

from ableton.v2.base import listens, listens_group, liveobj_valid
from ableton.v2.control_surface import Component, ParameterProvider
from ableton.v2.control_surface.control import ControlList, MappedSensitivitySettingControl
from ableton.v2.control_surface.elements import DisplayDataSource


class DeviceParameterComponent(Component):
    controls = ControlList(MappedSensitivitySettingControl, 8)

    def __init__(self, parameter_provider=None, *a, **k):
        (super().__init__)(*a, **k)
        self.parameter_provider = parameter_provider

    @property
    def parameter_provider(self):
        return self._parameter_provider

    @parameter_provider.setter
    def parameter_provider(self, provider):
        self._parameter_provider = provider or ParameterProvider()
        self._on_parameters_changed.subject = self._parameter_provider
        self._update_parameters()
        self._on_parameter_provider_changed(provider)

    def set_parameter_controls(self, encoders):
        self.controls.set_control_element(encoders)
        self._connect_parameters()

    def _connect_parameters(self):
        parameters = self._parameter_provider.parameters[: self.controls.control_count]
        for control, parameter_info in zip_longest(self.controls, parameters):
            parameter = parameter_info.parameter if parameter_info else None
            control.mapped_parameter = parameter
            if parameter:
                control.update_sensitivities(
                    parameter_info.default_encoder_sensitivity,
                    parameter_info.fine_grain_encoder_sensitivity,
                )

    def _update_parameters(self):
        if self.is_enabled():
            self._connect_parameters()

    @listens("parameters")
    def _on_parameters_changed(self):
        self._update_parameters()

    def _on_parameter_provider_changed(self, provider):
        pass

    def update(self):
        super().update()
        self._update_parameters()


class DisplayingDeviceParameterComponent(DeviceParameterComponent):
    def __init__(self, *a, **k):
        self._parameter_name_data_sources = list(map(DisplayDataSource, ("", "", "", "", "", "", "", "")))
        self._parameter_value_data_sources = list(map(DisplayDataSource, ("", "", "", "", "", "", "", "")))
        (super().__init__)(*a, **k)

    @property
    def parameters(self):
        return [p and p.parameter for p in self._parameter_provider.parameters]

    @property
    def parameter_names(self):
        return [p and (p.name) or "" for p in self.parameters]

    def set_parameter_name_displays(self, displays):
        for data_source, display in zip_longest(self._parameter_name_data_sources, displays or []):
            if display:
                display.set_data_sources([data_source])

    def set_parameter_value_displays(self, displays):
        for data_source, display in zip_longest(self._parameter_value_data_sources, displays or []):
            if display:
                display.set_data_sources([data_source])

    def set_name_display_line(self, line):
        self._set_display_line(line, self._parameter_name_data_sources)

    def set_value_display_line(self, line):
        self._set_display_line(line, self._parameter_value_data_sources)

    def _set_display_line(self, line, sources):
        if line:
            line.set_num_segments(len(sources))
            for segment in range(len(sources)):
                line.segment(segment).set_data_source(sources[segment])

    def clear_display(self):
        for source in chain(self._parameter_name_data_sources, self._parameter_value_data_sources):
            source.set_display_string("")

    def _update_parameters(self):
        super()._update_parameters()
        if self.is_enabled():
            parameters = self.parameters
            self._on_parameter_name_changed.replace_subjects(parameters)
            self._on_parameter_value_changed.replace_subjects(parameters)
            self._update_parameter_names()
            self._update_parameter_values()

    @listens_group("name")
    def _on_parameter_name_changed(self, parameter):
        self._update_parameter_names()

    @listens_group("value")
    def _on_parameter_value_changed(self, parameter):
        self._update_parameter_values()

    def _update_parameter_names(self):
        if self.is_enabled():
            params = zip(chain(self.parameter_provider.parameters, repeat(None)), self._parameter_name_data_sources)
            for info, name_data_source in params:
                name = self.info_to_name(info)
                name_data_source.set_display_string(name or "")

    def _update_parameter_values(self):
        if self.is_enabled():
            for parameter, data_source in zip_longest(self.parameters, self._parameter_value_data_sources):
                value_string = self.parameter_to_string(parameter)
                if data_source:
                    data_source.set_display_string(value_string)

    def info_to_name(self, info):
        return info and (info.name) or ""

    def parameter_to_string(self, parameter):
        if not liveobj_valid(parameter):
            return ""
        return unicode(parameter)

    def parameter_to_value(self, parameter):
        return parameter.value



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/elements/encoder.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 9830 bytes
import Live
from past.utils import old_div

from ...base import Event, clamp, const, in_range, listens, nop
from ..compound_element import CompoundElement
from ..input_control_element import MIDI_CC_TYPE, InputControlElement, InputSignal
from .combo import WrapperElement

_map_modes = map_modes = Live.MidiMap.MapMode
ABSOLUTE_MAP_MODES = (_map_modes.absolute, _map_modes.absolute_14_bit)


def _not_implemented(value):
    raise NotImplementedError


def signed_bit_delta(value):
    delta = SIGNED_BIT_DEFAULT_DELTA
    is_increment = value <= 64
    index = value - 1 if is_increment else value - 64
    if in_range(index, 0, len(SIGNED_BIT_VALUE_MAP)):
        delta = SIGNED_BIT_VALUE_MAP[index]
    if is_increment:
        return delta
    return -delta


SIGNED_BIT_DEFAULT_DELTA = 20.0
SIGNED_BIT_VALUE_MAP = (1, 2, 3, 4, 5, 8, 10, 20, 50)


def normalize_two_compliment(value):
    if value <= 64:
        return value
    return value - 128


ENCODER_VALUE_NORMALIZER = {
    _map_modes.relative_two_compliment: normalize_two_compliment,
    _map_modes.relative_smooth_two_compliment: normalize_two_compliment,
    _map_modes.relative_smooth_signed_bit: lambda v: v if v <= 64 else 64 - v,
    _map_modes.relative_smooth_binary_offset: lambda v: v - 64,
    _map_modes.relative_signed_bit: signed_bit_delta,
}
MAX_14_BIT_CC = 95


def accumulate_relative_two_compliment_chunk(chunk):
    right = 0
    left = 0
    for value in chunk:
        if value <= 64:
            right += value
        else:
            left += 128 - value

    result = clamp(right - left, -63, 64)
    if result >= 0:
        return result
    return 128 + result


ENCODER_VALUE_ACCUMULATOR = {_map_modes.relative_smooth_two_compliment: accumulate_relative_two_compliment_chunk}


class EncoderElement(InputControlElement):
    class ProxiedInterface(InputControlElement.ProxiedInterface):
        normalize_value = nop

    __events__ = (Event(name="normalized_value", signal=InputSignal),)
    encoder_sensitivity = 1.0
    allow_receiving_chunks = True

    def __init__(self, msg_type, channel, identifier, map_mode, encoder_sensitivity=None, *a, **k):
        (super().__init__)(msg_type, channel, identifier, *a, **k)
        if encoder_sensitivity is not None:
            self.encoder_sensitivity = encoder_sensitivity
        if map_mode is _map_modes.absolute_14_bit and identifier > MAX_14_BIT_CC:
            self._map_mode = _map_modes.absolute
        else:
            self._map_mode = map_mode
        self._last_received_value = None
        self._value_normalizer = ENCODER_VALUE_NORMALIZER.get(map_mode, _not_implemented)
        self._value_accumulator = ENCODER_VALUE_ACCUMULATOR.get(map_mode)
        self._half_value_range = self._max_value / 2

    def message_map_mode(self):
        return self._map_mode

    def receive_value(self, value):
        super().receive_value(value)
        self._last_received_value = value

    def relative_value_to_delta(self, value):
        if self._map_mode in ABSOLUTE_MAP_MODES:
            if self._last_received_value is not None:
                return value - self._last_received_value
            return 0
        return self._value_normalizer(value)

    def normalize_value(self, value):
        return old_div(self.relative_value_to_delta(value), self._half_value_range) * self.encoder_sensitivity

    def notify_value(self, value):
        super().notify_value(value)
        if self.normalized_value_listener_count():
            self.notify_normalized_value(self.normalize_value(value))

    def receive_chunk(self, chunk):
        if self._value_accumulator is not None:
            value = self._value_accumulator(chunk)
            if value != 0:
                self.receive_value(value)
        else:
            super().receive_chunk(chunk)


class TouchEncoderElementBase(EncoderElement):
    class ProxiedInterface(EncoderElement.ProxiedInterface):
        is_pressed = const(False)
        touch_element = const(None)

    __events__ = ("touch_value",)

    @property
    def touch_element(self):
        raise NotImplementedError

    def is_pressed(self):
        raise NotImplementedError


class TouchEncoderElement(CompoundElement, TouchEncoderElementBase):
    def __init__(
        self,
        msg_type=MIDI_CC_TYPE,
        channel=0,
        identifier=0,
        map_mode=_map_modes.absolute,
        touch_element=None,
        *a,
        **k,
    ):
        (super().__init__)(a, **k)
        self._touch_element = self.register_control_element(touch_element)
        self.request_listen_nested_control_elements()

    def on_nested_control_element_value(self, value, control):
        self.notify_touch_value(value)

    @property
    def touch_element(self):
        if self.owns_control_element(self._touch_element):
            return self._touch_element
        return None

    def is_pressed(self):
        return self.owns_control_element(self._touch_element) and self._touch_element.is_pressed()

    def on_nested_control_element_received(self, control):
        pass

    def on_nested_control_element_lost(self, control):
        pass


class FineGrainWithModifierEncoderElement(WrapperElement):
    def __init__(self, encoder=None, modifier=None, modified_sensitivity=0.1, default_sensitivity=None, *a, **k):
        (super().__init__)(a, wrapped_control=encoder, **k)
        self._normalized_value_listeners = []
        self._modified_sensitivity = modified_sensitivity
        self._default_sensitivity = default_sensitivity or self.wrapped_control.mapping_sensitivity
        self._modifier = modifier
        self.register_control_elements(modifier, encoder)
        self._FineGrainWithModifierEncoderElement__on_modifier_value.subject = self._modifier

    def add_normalized_value_listener(self, listener):
        self._normalized_value_listeners.append(listener)
        if len(self._normalized_value_listeners) == 1:
            self._enforce_control_invariant()

    def remove_normalized_value_listener(self, listener):
        self._normalized_value_listeners.remove(listener)
        if len(self._normalized_value_listeners) == 0:
            self._enforce_control_invariant()

    def normalized_value_has_listener(self, listener):
        return listener in self._normalized_value_listeners

    @listens("normalized_value")
    def __on_normalized_value(self, value):
        for listener in self._normalized_value_listeners:
            listener(value)

    @listens("value")
    def __on_modifier_value(self, value):
        if self.owns_control_element(self._modifier):
            self.on_nested_control_element_value(value, self._modifier)

    def on_nested_control_element_received(self, control):
        super().on_nested_control_element_received(control)
        self._enforce_control_invariant()

    def on_nested_control_element_lost(self, control):
        super().on_nested_control_element_lost(control)
        self._enforce_control_invariant()

    def on_nested_control_element_value(self, value, control):
        if control == self._modifier:
            self._enforce_control_invariant()
        else:
            super().on_nested_control_element_value(value, control)

    def _enforce_control_invariant(self):
        if self.owns_control_element(self._wrapped_control):
            if self._modifier.is_pressed():
                self.wrapped_control.mapping_sensitivity = self._modified_sensitivity
            else:
                self.wrapped_control.mapping_sensitivity = self._default_sensitivity
        should_listen = self.owns_control_element(self._wrapped_control) and len(self._normalized_value_listeners) > 0
        self._FineGrainWithModifierEncoderElement__on_normalized_value.subject = (
            self._wrapped_control if should_listen else None
        )

    def set_sensitivities(self, default, modified):
        self._default_sensitivity = default
        self._modified_sensitivity = modified
        self._enforce_control_invariant()



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/_Framework/Dependency.py
# Compiled at: 2023-12-21 15:35:35
# Size of source mod 2**32: 5364 bytes
__all__ = ("inject", "depends", "dependency")
from functools import wraps

from .Util import union


class DependencyError(Exception):
    pass


class InjectionRegistry:
    def __init__(self, parent=None, *a, **k):
        (super().__init__)(*a, **k)
        self._key_registry = {}

    def register_key(self, key, injector):
        self._key_registry.setdefault(key, []).append(injector)

    def unregister_key(self, key, injector):
        self._key_registry[key].remove(injector)
        if not self._key_registry[key]:
            del self._key_registry[key]

    def get(self, key, default=None):
        try:
            return self._key_registry[key][-1].provides[key]
        except KeyError:
            return default


_global_injection_registry = InjectionRegistry()


def get_dependency_for(obj, name, default=None):
    accessor = _global_injection_registry.get(name, default)
    if accessor is not None:
        return accessor()
    raise DependencyError(f"Required dependency {name} not provided for {obj!s}")


class dependency:
    def __init__(self, **k):
        self._dependency_name, self._dependency_default = next(iter(k.items()))

    def __get__(self, obj, cls=None):
        if obj is None:
            obj = cls
        return get_dependency_for(obj, self._dependency_name, self._dependency_default)


def depends(**dependencies):
    def decorator(func):
        @wraps(func)
        def wrapper(self, *a, **explicit):
            deps = {k: get_dependency_for(self, k, v) for k, v in dependencies.items() if k not in explicit}
            return func(self, *a, **union(deps, explicit))

        return wrapper

    return decorator


class Injector:
    @property
    def provides(self):
        return {}

    def register(self):
        pass

    def unregister(self):
        pass

    def __enter__(self):
        self.register()
        return self

    def __exit__(self, *a):
        self.unregister()


class RegistryInjector(Injector):
    def __init__(self, provides=None, registry=None, *a, **k):
        (super(Injector, self).__init__)(*a, **k)
        self._provides_dict = provides
        self._registry = registry

    @property
    def provides(self):
        return self._provides_dict

    def register(self):
        registry = self._registry
        for k in self._provides_dict:
            registry.register_key(k, self)

    def unregister(self):
        registry = self._registry
        for k in self._provides_dict:
            registry.unregister_key(k, self)


class InjectionFactory:
    def __init__(self, provides=None, *a, **k):
        (super().__init__)(*a, **k)
        self._provides_dict = provides

    def everywhere(self):
        return RegistryInjector(provides=(self._provides_dict), registry=_global_injection_registry)

    into_object = NotImplemented
    into_class = NotImplemented


def inject(**k):
    return InjectionFactory(k)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/_Framework/Skin.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 1389 bytes
from itertools import chain

from future.utils import raise_


class SkinColorMissingError(Exception):
    pass


class Skin:
    def __init__(self, colors=None, *a, **k):
        (super().__init__)(*a, **k)
        self._colors = {}
        if colors is not None:
            self._fill_colors(colors)

    def _fill_colors(self, colors, pathname=""):
        if getattr(colors, "__bases__", None):
            for base in colors.__bases__:
                self._fill_colors(base)

        for k, v in colors.__dict__.items():
            if k[:1] != "_":
                if callable(v):
                    self._fill_colors(v, pathname + k + ".")
                else:
                    self._colors[pathname + k] = v

    def __getitem__(self, key):
        try:
            return self._colors[key]
        except KeyError:
            raise_(SkinColorMissingError, f"Skin color missing: {key!s}")

    def iteritems(self):
        return iter(self._colors.items())


def merge_skins(*skins):
    skin = Skin()
    skin._colors = dict(chain(*[list(s._colors.items()) for s in skins]))
    return skin



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/delay_decoration.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 4374 bytes
from ableton.v2.base import EventObject, listens, liveobj_valid
from ableton.v2.control_surface import LiveObjectDecorator, NotifyingList, get_parameter_by_name

from .internal_parameter import EnumWrappingParameter


class ChannelType(int):
    pass


ChannelType.left_right = ChannelType(0)
ChannelType.left = ChannelType(1)
ChannelType.right = ChannelType(2)


class LinkedState(int):
    pass


LinkedState.unlinked = LinkedState(0)
LinkedState.linked = LinkedState(1)


class DelayDeviceDecorator(LiveObjectDecorator, EventObject):
    sync_modes = ("Time", "Sync")
    link_modes = ("Linked", "Unlinked")

    def __init__(self, *a, **k):
        (super().__init__)(*a, **k)
        self._channel_type_provider = NotifyingList(
            available_values=["L+R", "Left", "Right"],
            default_value=(ChannelType.left_right),
        )
        self._additional_parameters = self._create_parameters()
        self.register_disconnectables(self._additional_parameters)

    @property
    def parameters(self):
        return tuple(self._live_object.parameters) + self._additional_parameters

    def _create_parameters(self):
        self._channel_switch = EnumWrappingParameter(
            name="Channel",
            parent=self,
            values_host=(self._channel_type_provider),
            index_property_host=(self._channel_type_provider),
            values_property="available_values",
            index_property="index",
            value_type=ChannelType,
        )
        self._delay_line_link_parameter = get_parameter_by_name(self, "Link")
        self._DelayDeviceDecorator__on_channel_switch_changed.subject = self._channel_switch
        self._DelayDeviceDecorator__on_linked_changed.subject = self._delay_line_link_parameter
        self._DelayDeviceDecorator__on_linked_changed()
        return (
            self._channel_switch,
            EnumWrappingParameter(
                name="L Sync Enum",
                parent=self,
                values_host=self,
                values_property="sync_modes",
                index_property_host=(get_parameter_by_name(self, "L Sync")),
                index_property="value",
                to_index_conversion=(lambda index: int(index)),
                from_index_conversion=(lambda index: int(index)),
            ),
            EnumWrappingParameter(
                name="R Sync Enum",
                parent=self,
                values_host=self,
                values_property="sync_modes",
                index_property_host=(get_parameter_by_name(self, "R Sync")),
                index_property="value",
                to_index_conversion=(lambda index: int(index)),
                from_index_conversion=(lambda index: int(index)),
            ),
            EnumWrappingParameter(
                name="Link Switch",
                parent=self,
                values_host=self,
                values_property="link_modes",
                index_property_host=(self._delay_line_link_parameter),
                index_property="value",
                to_index_conversion=(lambda index: int(not index)),
                from_index_conversion=(lambda index: int(not index)),
            ),
        )

    def _linked_state_needs_updating(self):
        if liveobj_valid(self._delay_line_link_parameter):
            is_delay_line_linked = self._delay_line_link_parameter.value == LinkedState.linked
            is_channel_switch_linked = self._channel_switch.value == ChannelType.left_right
            return is_delay_line_linked != is_channel_switch_linked
        return None

    @listens("value")
    def __on_channel_switch_changed(self):
        if self._linked_state_needs_updating():
            self._delay_line_link_parameter.value = (
                LinkedState.linked if self._channel_switch.value == ChannelType.left_right else LinkedState.unlinked
            )

    @listens("value")
    def __on_linked_changed(self):
        if self._linked_state_needs_updating():
            self._channel_switch.value = (
                ChannelType.left_right
                if self._delay_line_link_parameter.value == LinkedState.linked
                else ChannelType.left
            )



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/components/auto_arm.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 3482 bytes
from ...base import depends, listens, listens_group, task
from ...live import any_track_armed, liveobj_changed, liveobj_valid
from .. import Component


def track_can_be_auto_armed(track):
    return liveobj_valid(track) and track.can_be_armed and track.has_midi_input


class AutoArmComponent(Component):
    @depends(target_track=None)
    def __init__(self, name="Auto_Arm", target_track=None, *a, **k):
        (super().__init__)(a, name=name, **k)
        self._target_track = target_track
        self._auto_arm_target = None
        self._update_auto_arm_task = self._tasks.add(task.run(self._update_auto_arm))
        self.register_slot(self._target_track, self.update, "target_track")
        self._AutoArmComponent__on_tracks_changed.subject = self.song
        self._AutoArmComponent__on_tracks_changed()

    def disconnect(self):
        self._setup_new_auto_arm_target(None)
        super().disconnect()

    def update(self):
        super().update()
        self._update_auto_arm()

    def _can_auto_arm(self):
        return self.is_enabled() and not any_track_armed()

    def _auto_arm_target_changed(self, target_track):
        return liveobj_changed(target_track, self._auto_arm_target) or not track_can_be_auto_armed(
            self._auto_arm_target,
        )

    def _set_auto_arm_state(self, state):
        if liveobj_valid(self._auto_arm_target):
            if self._auto_arm_target.implicit_arm != state:
                self._auto_arm_target.implicit_arm = state

    def _setup_new_auto_arm_target(self, target_track):
        new_target = target_track if track_can_be_auto_armed(target_track) else None
        self._AutoArmComponent__on_implicit_arm_changed.subject = new_target
        self._set_auto_arm_state(False)
        self._auto_arm_target = new_target

    def _update_auto_arm(self):
        self._update_auto_arm_task.kill()
        if self._can_auto_arm():
            target_track = self._target_track.target_track
            if self._auto_arm_target_changed(target_track):
                self._setup_new_auto_arm_target(target_track)
            self._set_auto_arm_state(True)
        else:
            self._setup_new_auto_arm_target(None)

    @listens("implicit_arm")
    def __on_implicit_arm_changed(self):
        self._update_auto_arm()

    @listens_group("arm")
    def __on_arm_changed(self, _):
        self._update_auto_arm_task.restart()

    @listens_group("input_routing_type")
    def __on_input_routing_type_changed(self, _):
        self._update_auto_arm_task.restart()

    @listens_group("is_frozen")
    def __on_frozen_state_changed(self, _):
        self._update_auto_arm_task.restart()

    @listens("tracks")
    def __on_tracks_changed(self):
        tracks = list(filter(lambda t: t.can_be_armed, self.song.tracks))
        self._AutoArmComponent__on_arm_changed.replace_subjects(tracks)
        self._AutoArmComponent__on_input_routing_type_changed.replace_subjects(tracks)
        self._AutoArmComponent__on_frozen_state_changed.replace_subjects(tracks)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/components/__init__.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 3382 bytes
from .accent import AccentComponent
from .auto_arm import AutoArmBase, AutoArmComponent
from .background import BackgroundComponent, ModifierBackgroundComponent
from .channel_strip import ChannelStripComponent
from .clip_actions import ClipActionsComponent
from .clip_slot import ClipSlotComponent, find_nearest_color
from .device import DeviceComponent
from .device_navigation import DeviceNavigationComponent, FlattenedDeviceChain, is_empty_rack, nested_device_parent
from .device_parameters import DeviceParameterComponent, DisplayingDeviceParameterComponent
from .drum_group import DrumGroupComponent
from .item_lister import ItemListerComponent, ItemProvider, ItemSlot, SimpleItemSlot
from .mixer import MixerComponent, RightAlignTracksTrackAssigner, SimpleTrackAssigner
from .playable import PlayableComponent
from .scene import SceneComponent
from .scroll import Scrollable, ScrollComponent
from .session import SessionComponent
from .session_navigation import (
    SessionNavigationComponent,
    SessionRingScenePager,
    SessionRingSceneScroller,
    SessionRingScroller,
    SessionRingTrackPager,
    SessionRingTrackScroller,
)
from .session_overview import SessionOverviewComponent
from .session_recording import SessionRecordingComponent, track_is_recording, track_playing_slot
from .session_ring import SessionRingComponent
from .slide import Slideable, SlideComponent
from .target_track import ArmedTargetTrackComponent, TargetTrackComponent
from .toggle import ToggleComponent
from .transport import TransportComponent
from .undo_redo import UndoRedoComponent
from .view_control import (
    BasicSceneScroller,
    BasicTrackScroller,
    SceneListScroller,
    SceneScroller,
    TrackScroller,
    ViewControlComponent,
    all_tracks,
)

__all__ = (
    "AccentComponent",
    "all_tracks",
    "ArmedTargetTrackComponent",
    "AutoArmBase",
    "AutoArmComponent",
    "BackgroundComponent",
    "ModifierBackgroundComponent",
    "ChannelStripComponent",
    "ClipActionsComponent",
    "ClipSlotComponent",
    "find_nearest_color",
    "DeviceComponent",
    "DeviceNavigationComponent",
    "DeviceParameterComponent",
    "DisplayingDeviceParameterComponent",
    "DrumGroupComponent",
    "FlattenedDeviceChain",
    "is_empty_rack",
    "ItemListerComponent",
    "ItemProvider",
    "ItemSlot",
    "MixerComponent",
    "nested_device_parent",
    "PlayableComponent",
    "RightAlignTracksTrackAssigner",
    "SceneComponent",
    "Scrollable",
    "ScrollComponent",
    "SessionComponent",
    "SessionNavigationComponent",
    "SessionRingScroller",
    "SessionRingTrackScroller",
    "SessionRingSceneScroller",
    "SessionRingTrackPager",
    "SessionRingScenePager",
    "SessionRecordingComponent",
    "SessionRingComponent",
    "SessionOverviewComponent",
    "SimpleItemSlot",
    "SimpleTrackAssigner",
    "Slideable",
    "SlideComponent",
    "TargetTrackComponent",
    "ToggleComponent",
    "TransportComponent",
    "BasicSceneScroller",
    "BasicTrackScroller",
    "SceneListScroller",
    "SceneScroller",
    "track_is_recording",
    "track_playing_slot",
    "TrackScroller",
    "UndoRedoComponent",
    "ViewControlComponent",
)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/components/session_overview.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 7345 bytes
from past.utils import old_div

from ...base import in_range, listens, liveobj_valid
from ..component import Component


class SessionOverviewComponent(Component):
    def __init__(self, session_ring=None, enable_skinning=False, *a, **k):
        (super().__init__)(*a, **k)
        self._buttons = None
        self._scene_bank_index = 0
        self._track_bank_index = 0
        self._empty_value = 0
        self._stopped_value = 100
        self._playing_value = 127
        self._selected_value = 64
        self._session_ring = session_ring
        self._SessionOverviewComponent__on_session_offset_changes.subject = self._session_ring
        self._SessionOverviewComponent__on_scene_list_changed.subject = self.song
        if enable_skinning:
            self._enable_skinning()

    def _enable_skinning(self):
        self.set_stopped_value("Zooming.Stopped")
        self.set_selected_value("Zooming.Selected")
        self.set_playing_value("Zooming.Playing")
        self.set_empty_value("Zooming.Empty")

    @listens("scenes")
    def __on_scene_list_changed(self):
        self.update()

    def set_button_matrix(self, buttons):
        if buttons:
            buttons.reset()
        self._buttons = buttons
        self._SessionOverviewComponent__on_matrix_value.subject = self._buttons
        self.update()
        self._update_bank_index(self._session_ring.track_offset, self._session_ring.scene_offset)

    def set_empty_value(self, value):
        self._empty_value = value

    def set_playing_value(self, value):
        self._playing_value = value

    def set_stopped_value(self, value):
        self._stopped_value = value

    def set_selected_value(self, value):
        self._selected_value = value

    def update(self):
        super().update()
        if self.is_enabled():
            self._update_matrix_buttons()

    def _update_matrix_buttons(self):
        if self._buttons is not None:
            tracks = self._session_ring.tracks_to_use()
            scenes = self.song.scenes
            slots_registry = [None for index in range(len(scenes))]
            width = self._session_ring.num_tracks
            height = self._session_ring.num_scenes
            for x in range(self._buttons.width()):
                for y in range(self._buttons.height()):
                    value_to_send = self._empty_value
                    track_bank_offset = self._track_bank_index * self._buttons.width() * width
                    scene_bank_offset = self._scene_bank_index * self._buttons.height() * height
                    track_offset = x * width + track_bank_offset
                    scene_offset = y * height + scene_bank_offset
                    if track_offset in range(len(tracks)):
                        if scene_offset in range(len(scenes)):
                            value_to_send = self._stopped_value
                            if self._session_ring.track_offset - track_bank_offset in range(
                                width * (x - 1) + 1,
                                width * (x + 1),
                            ) and self._session_ring.scene_offset - scene_bank_offset in range(
                                height * (y - 1) + 1,
                                height * (y + 1),
                            ):
                                value_to_send = self._selected_value
                            else:
                                playing = False
                                for track in range(track_offset, track_offset + width):
                                    for scene in range(scene_offset, scene_offset + height):
                                        if track in range(len(tracks)):
                                            if scene in range(len(scenes)):
                                                if not liveobj_valid(slots_registry[scene]):
                                                    slots_registry[scene] = scenes[scene].clip_slots
                                                else:
                                                    slot = (
                                                        slots_registry[scene][track]
                                                        if len(slots_registry[scene]) > track
                                                        else None
                                                    )
                                                if liveobj_valid(slot):
                                                    if slot.has_clip:
                                                        if slot.clip.is_playing:
                                                            value_to_send = self._playing_value
                                                            playing = True
                                                            break

                                    if playing:
                                        break

                    if in_range(value_to_send, 0, 128):
                        self._buttons.send_value(x, y, value_to_send)
                    else:
                        self._buttons.set_light(x, y, value_to_send)

    @listens("offset")
    def __on_session_offset_changes(self, track_offset, scene_offset):
        self._update_bank_index(track_offset, scene_offset)

    def _update_bank_index(self, track_offset, scene_offset):
        if self._buttons:
            if self.is_enabled():
                self._track_bank_index = int(
                    old_div(old_div(track_offset, self._session_ring.num_tracks), self._buttons.width()),
                )
                self._scene_bank_index = int(
                    old_div(old_div(scene_offset, self._session_ring.num_scenes), self._buttons.height()),
                )
        self.update()

    @listens("value")
    def __on_matrix_value(self, value, x, y, is_momentary):
        if self.is_enabled():
            if not (value != 0 or is_momentary):
                track_offset = (x + self._track_bank_index * self._buttons.width()) * self._session_ring.num_tracks
                scene_offset = (y + self._scene_bank_index * self._buttons.height()) * self._session_ring.num_scenes
                if track_offset in range(len(self._session_ring.tracks_to_use())):
                    if scene_offset in range(len(self.song.scenes)):
                        self._session_ring.set_offsets(track_offset, scene_offset)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/_Framework/ButtonElement.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 3778 bytes
import Live

from .InputControlElement import InputControlElement
from .Skin import Skin, SkinColorMissingError
from .Util import nop


class ButtonValue:
    midi_value = 0

    def __init__(self, midi_value=None, *a, **k):
        (super().__init__)(*a, **k)
        if midi_value is not None:
            self.midi_value = midi_value

    def __int__(self):
        return self.midi_value

    def __eq__(self, other):
        try:
            return id(self) == id(other) or self.midi_value == other
        except NotImplementedError:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __bool__(self):
        return self != OFF_VALUE


ON_VALUE = ButtonValue(127)
OFF_VALUE = ButtonValue(0)


class Color(ButtonValue):
    def draw(self, interface):
        interface.send_value(self.midi_value)


class DummyUndoStepHandler:
    def begin_undo_step(self):
        pass

    def end_undo_step(self):
        pass


class ButtonElementMixin:
    def set_light(self, is_turned_on):
        if is_turned_on:
            self.turn_on()
        else:
            self.turn_off()

    def turn_on(self):
        self.send_value(ON_VALUE)

    def turn_off(self):
        self.send_value(OFF_VALUE)


class ButtonElement(InputControlElement, ButtonElementMixin):
    class ProxiedInterface(InputControlElement.ProxiedInterface, ButtonElementMixin):
        is_momentary = nop
        is_pressed = nop

    def __init__(
        self,
        is_momentary,
        msg_type,
        channel,
        identifier,
        skin=Skin(),
        undo_step_handler=DummyUndoStepHandler(),
        *a,
        **k,
    ):
        (super().__init__)(msg_type, channel, identifier, *a, **k)
        self._ButtonElement__is_momentary = bool(is_momentary)
        self._last_received_value = -1
        self._undo_step_handler = undo_step_handler
        self._skin = skin

    def is_momentary(self):
        return self._ButtonElement__is_momentary

    def message_map_mode(self):
        return Live.MidiMap.MapMode.absolute

    def is_pressed(self):
        return self._ButtonElement__is_momentary and int(self._last_received_value) > 0

    def set_light(self, value):
        self._set_skin_light(value)

    def _set_skin_light(self, value):
        try:
            color = self._skin[value]
            color.draw(self)
        except SkinColorMissingError:
            super().set_light(value)

    def receive_value(self, value):
        pressed_before = self.is_pressed()
        self._last_received_value = value
        if not pressed_before:
            if self.is_pressed():
                self._undo_step_handler.begin_undo_step()
        super().receive_value(value)
        if pressed_before:
            if not self.is_pressed():
                self._undo_step_handler.end_undo_step()

    def disconnect(self):
        super().disconnect()
        self._undo_step_handler = None



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/base/disconnectable.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 2483 bytes
from .util import find_if


class Disconnectable:
    def disconnect(self):
        pass


class CompoundDisconnectable(Disconnectable):
    def __init__(self, *a, **k):
        (super().__init__)(*a, **k)
        self._registered_disconnectables = []

    def register_disconnectables(self, disconnectables):
        for disconnectable in disconnectables:
            self.register_disconnectable(disconnectable)

        return disconnectables

    def register_disconnectable(self, slot):
        if slot not in self._registered_disconnectables:
            self._registered_disconnectables.append(slot)
        return slot

    def unregister_disconnectable(self, slot):
        if slot in self._registered_disconnectables:
            self._registered_disconnectables.remove(slot)

    def disconnect_disconnectable(self, slot):
        if slot in self._registered_disconnectables:
            self._registered_disconnectables.remove(slot)
            slot.disconnect()

    def find_disconnectable(self, predicate):
        return find_if(predicate, self._registered_disconnectables)

    def has_disconnectable(self, slot):
        return slot in self._registered_disconnectables

    def disconnect(self):
        for slot in self._registered_disconnectables:
            slot.disconnect()

        self._registered_disconnectables = []
        super().disconnect()


class disconnectable:
    def __init__(self, managed=None, *a, **k):
        (super().__init__)(*a, **k)
        self._managed = managed

    def __enter__(self):
        return self._managed

    def __exit__(self, *a, **k):
        if self._managed is not None:
            self._managed.disconnect()



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/_Generic/util.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 1644 bytes
from _Framework.SubjectSlot import SlotManager, subject_slot
from _Framework.Util import nop


class DeviceAppointer(SlotManager):
    def __init__(self, song=None, appointed_device_setter=nop, *a, **k):
        (super().__init__)(*a, **k)
        self._set_appointed_device = appointed_device_setter
        self._appointed_device = None
        self._song = song
        self._DeviceAppointer__on_appointed_device_changed.subject = self._song
        self._DeviceAppointer__on_selected_track_changed.subject = self._song.view
        self._DeviceAppointer__on_selected_track_changed()

    @subject_slot("appointed_device")
    def __on_appointed_device_changed(self):
        if self._appointed_device != self._song.appointed_device:
            self._update_appointed_device(self._song.appointed_device)

    @subject_slot("selected_device")
    def __on_selected_device_changed(self):
        song = self._song
        device = song.view.selected_track.view.selected_device
        if device is not None:
            self._update_appointed_device(device)

    @subject_slot("selected_track")
    def __on_selected_track_changed(self):
        track_view = self._song.view.selected_track.view
        self._DeviceAppointer__on_selected_device_changed.subject = track_view
        self._update_appointed_device(track_view.selected_device)

    def _update_appointed_device(self, device):
        if device is not None:
            self._appointed_device = device
            self._set_appointed_device(device)
            self._song.appointed_device = device



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/_Framework/ComboElement.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 14812 bytes
from contextlib import contextmanager

from ableton.v2.base import PY3

from . import Defaults, Task
from .ButtonElement import ButtonElementMixin
from .CompoundElement import CompoundElement
from .Dependency import depends
from .InputControlElement import ParameterSlot
from .NotifyingControlElement import NotifyingControlElement
from .Proxy import ProxyBase
from .Resource import DEFAULT_PRIORITY
from .SubjectSlot import SlotManager, Subject, subject_slot
from .Util import const, find_if, lazy_attribute, nop


class WrapperElement(CompoundElement, ProxyBase):
    class ProxiedInterface(CompoundElement.ProxiedInterface):
        def __getattr__(self, name):
            if PY3:
                if "_wrapped_control" not in self.outer.__dict__:
                    raise AttributeError()
            wrapped = self.outer.__dict__["_wrapped_control"]
            return getattr(wrapped.proxied_interface, name)

    def __init__(self, wrapped_control=None, *a, **k):
        (super().__init__)(*a, **k)
        self._wrapped_control = wrapped_control
        self._parameter_slot = ParameterSlot()

    @property
    def proxied_object(self):
        if self._is_initialized():
            if self.owns_control_element(self._wrapped_control):
                return self._wrapped_control
            return None
        return None

    def _is_initialized(self):
        return "_wrapped_control" in self.__dict__

    def register_wrapped(self):
        self.register_control_element(self._wrapped_control)

    def unregister_wrapped(self):
        self.unregister_control_element(self._wrapped_control)

    @property
    def wrapped_control(self):
        return self._wrapped_control

    def __bool__(self):
        return self.owns_control_element(self._wrapped_control)

    def on_nested_control_element_received(self, control):
        if control == self._wrapped_control:
            self._parameter_slot.control = control

    def on_nested_control_element_lost(self, control):
        if control == self._wrapped_control:
            self._parameter_slot.control = None

    def on_nested_control_element_value(self, value, control):
        if control == self._wrapped_control:
            self.notify_value(value)

    def connect_to(self, parameter):
        if self._parameter_slot.parameter is None:
            self.request_listen_nested_control_elements()
        self._parameter_slot.parameter = parameter

    def release_parameter(self):
        if self._parameter_slot.parameter is not None:
            self.unrequest_listen_nested_control_elements()
        self._parameter_slot.parameter = None


class ComboElement(WrapperElement):
    priority_increment = 0.5

    def __init__(self, control=None, modifiers=None, negative_modifiers=None, *a, **k):
        if negative_modifiers is None:
            negative_modifiers = []
        if modifiers is None:
            modifiers = []
        (super().__init__)(a, wrapped_control=control, **k)
        self._combo_modifiers = dict([(x, True) for x in modifiers] + [(x, False) for x in negative_modifiers])
        (self.register_control_elements)(*list(self._combo_modifiers.keys()))
        self.request_listen_nested_control_elements()

    def reset(self):
        if self.owns_control_element(self._wrapped_control):
            self._wrapped_control.reset()

    def get_control_element_priority(self, element, priority):
        if element == self._wrapped_control:
            priority = DEFAULT_PRIORITY if priority is None else priority
            return priority + self.priority_increment
        return priority

    def on_nested_control_element_received(self, control):
        if control != self._wrapped_control:
            self._enforce_control_invariant()
        else:
            super().on_nested_control_element_received(control)

    def on_nested_control_element_lost(self, control):
        if control != self._wrapped_control:
            self._enforce_control_invariant()
        else:
            super().on_nested_control_element_lost(control)

    def on_nested_control_element_value(self, value, control):
        if control != self._wrapped_control:
            self._enforce_control_invariant()
        else:
            super().on_nested_control_element_value(value, control)

    def _enforce_control_invariant(self):
        if self._combo_is_on():
            if not self.has_control_element(self._wrapped_control):
                self.register_control_element(self._wrapped_control)
        else:
            if self.has_control_element(self._wrapped_control):
                self.unregister_control_element(self._wrapped_control)

    def _combo_is_on(self):
        return all(map(self._modifier_is_valid, self._combo_modifiers))

    def _modifier_is_valid(self, mod):
        return self.owns_control_element(mod) and mod.is_pressed() == self._combo_modifiers[mod]


class EventElement(NotifyingControlElement, SlotManager, ProxyBase, ButtonElementMixin):
    event_value = 1
    _subject = None

    def __init__(self, subject=None, event=None, *a, **k):
        (super().__init__)(*a, **k)
        self._subject = subject
        self.register_slot(subject, self._on_event, event)

    @property
    def proxied_object(self):
        return self._subject

    @property
    def proxied_interface(self):
        return getattr(self._subject, "proxied_interface", self._subject)

    def _on_event(self, *a, **k):
        self.notify_value(self.event_value)

    def is_momentary(self):
        return False

    def reset(self):
        getattr(self._subject, "reset", nop)()

    def send_value(self, *a, **k):
        try:
            send_value = super().__getattr__("send_value")
        except AttributeError:
            send_value = nop

        send_value(*a, **k)

    def set_light(self, *a, **k):
        try:
            set_light = super().__getattr__("set_light")
        except AttributeError:
            set_light = nop

        set_light(*a, **k)


class DoublePressContext(Subject):
    __subject_events__ = ("break_double_press",)

    @contextmanager
    def breaking_double_press(self):
        self._broke_double_press = False
        yield
        if not self._broke_double_press:
            self.break_double_press()

    def break_double_press(self):
        self.notify_break_double_press()
        self._broke_double_press = True


GLOBAL_DOUBLE_PRESS_CONTEXT_PROVIDER = const(DoublePressContext())


class DoublePressElement(WrapperElement):
    __subject_events__ = ("single_press", "double_press")
    DOUBLE_PRESS_MAX_DELAY = Defaults.MOMENTARY_DELAY

    @depends(double_press_context=GLOBAL_DOUBLE_PRESS_CONTEXT_PROVIDER)
    def __init__(self, wrapped_control=None, double_press_context=None, *a, **k):
        (super().__init__)(a, wrapped_control=wrapped_control, **k)
        self.register_control_element(self._wrapped_control)
        self._double_press_context = double_press_context
        self._double_press_task = self._tasks.add(
            Task.sequence(Task.wait(self.DOUBLE_PRESS_MAX_DELAY), Task.run(self.finish_single_press)),
        ).kill()
        self.request_listen_nested_control_elements()

    def on_nested_control_element_value(self, value, control):
        if not control.is_momentary() or value and self._double_press_task.is_killed:
            self._double_press_context.break_double_press()
            self._on_break_double_press.subject = self._double_press_context
            self._double_press_task.restart()
        else:
            self.finish_double_press()
            self._double_press_task.kill()
        super().on_nested_control_element_value(value, control)

    @subject_slot("break_double_press")
    def _on_break_double_press(self):
        if not self._double_press_task.is_killed:
            self._double_press_task.kill()
            self.finish_single_press()

    def finish_single_press(self):
        self._on_break_double_press.subject = None
        self.notify_single_press()

    def finish_double_press(self):
        self._on_break_double_press.subject = None
        self.notify_double_press()

    @lazy_attribute
    def single_press(self):
        return EventElement(self, "single_press")

    @lazy_attribute
    def double_press(self):
        return EventElement(self, "double_press")


class MultiElement(CompoundElement, ButtonElementMixin):
    class ProxiedInterface(CompoundElement.ProxiedInterface):
        def __getattr__(self, name):
            found = find_if(
                lambda x: x is not None,
                (getattr(c.proxied_interface, name, None) for c in self.outer.nested_control_elements()),
            )
            if found is not None:
                return found
            raise AttributeError

    def __init__(self, *controls, **k):
        (super().__init__)(**k)
        (self.register_control_elements)(*controls)

    def send_value(self, value):
        for control in self.owned_control_elements():
            control.send_value(value)

    def set_light(self, value):
        for control in self.owned_control_elements():
            control.set_light(value)

    def on_nested_control_element_value(self, value, control):
        if not self.is_pressed() or value:
            self.notify_value(value)

    def is_pressed(self):
        return find_if(lambda c: getattr(c, "is_pressed", const(False))(), self.owned_control_elements()) is not None

    def is_momentary(self):
        return find_if(lambda c: getattr(c, "is_momentary", const(False))(), self.nested_control_elements()) is not None

    def on_nested_control_element_received(self, control):
        pass

    def on_nested_control_element_lost(self, control):
        pass


class ToggleElement(WrapperElement):
    def __init__(self, on_control=None, off_control=None, *a, **k):
        (super().__init__)(*a, **k)
        self._on_control = on_control
        self._off_control = off_control
        self._toggled = False
        self._update_toggled()

    def set_toggled(self, value):
        self._toggled = value
        self._update_toggled()

    def _update_toggled(self):
        if self.has_control_element(self._wrapped_control):
            self.unregister_control_element(self._wrapped_control)
        self._wrapped_control = self._on_control if self._toggled else self._off_control
        if self._wrapped_control is not None:
            self.register_control_element(self._wrapped_control)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/mode/__init__.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 1432 bytes
from ableton.v2.control_surface.mode import (
    AddLayerMode,
    CompoundMode,
    EnablingMode,
    ImmediateBehaviour,
    LatchingBehaviour,
    LayerMode,
    LayerModeBase,
    Mode,
    ModeButtonBehaviour,
    ModeButtonControl,
    make_mode_button_control,
    pop_last_mode,
)

from .behaviour import MomentaryBehaviour, ReenterBehaviourMixin, ToggleBehaviour, make_reenter_behaviour
from .mode import CallFunctionMode, EnablingAddLayerMode, ShowDetailClipMode
from .modes import ModesComponent
from .selector import (
    EventDescription,
    select_mode_for_main_view,
    select_mode_on_event_change,
    toggle_mode_on_property_change,
)

__all__ = (
    "AddLayerMode",
    "CallFunctionMode",
    "CompoundMode",
    "EnablingAddLayerMode",
    "EnablingMode",
    "EventDescription",
    "ImmediateBehaviour",
    "LatchingBehaviour",
    "LayerMode",
    "LayerModeBase",
    "Mode",
    "ModeButtonControl",
    "ModeButtonBehaviour",
    "ModesComponent",
    "MomentaryBehaviour",
    "ReenterBehaviourMixin",
    "ShowDetailClipMode",
    "ToggleBehaviour",
    "make_mode_button_control",
    "make_reenter_behaviour",
    "pop_last_mode",
    "select_mode_for_main_view",
    "select_mode_on_event_change",
    "toggle_mode_on_property_change",
)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/banking_util.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 6851 bytes
from math import ceil

from ableton.v2.control_surface import BankingInfo as BankingInfoBase
from ableton.v2.control_surface import DescribedDeviceParameterBank as DescribedDeviceParameterBankBase
from ableton.v2.control_surface.device_parameter_bank import DeviceParameterBank
from ableton.v2.control_surface.device_parameter_bank import MaxDeviceParameterBank as MaxDeviceParameterBankBase

from ..live import liveobj_valid
from . import BANK_FORMAT, BANK_PARAMETERS_KEY, all_parameters

BANK_NAME_JOIN_STR = " and "
DEFAULT_BANK_SIZE = 8


def create_parameter_bank(device, banking_info):
    bank = None
    if liveobj_valid(device):
        bank_class = DeviceParameterBank
        size = banking_info.bank_size
        if size >= DEFAULT_BANK_SIZE:
            if banking_info.has_bank_count(device):
                bank_class = MaxDeviceParameterBank
            else:
                if banking_info.device_bank_definition(device) is not None:
                    bank_class = DescribedDeviceParameterBank
        bank = bank_class(device=device, size=size, banking_info=banking_info)
    return bank


class DescribedDeviceParameterBank(DescribedDeviceParameterBankBase):
    def _current_parameter_slots(self):
        next_index = self.index + 1
        bank_count = self._banking_info.device_bank_count(self._device)
        if self._size > DEFAULT_BANK_SIZE:
            if next_index < bank_count:
                return (self._definition.value_by_index(self.index).get(BANK_PARAMETERS_KEY) or ()) + (
                    self._definition.value_by_index(next_index).get(BANK_PARAMETERS_KEY) or ()
                )
        return super()._current_parameter_slots()

    def _calc_name(self):
        if self._size > DEFAULT_BANK_SIZE:
            return self._banking_info.device_bank_names(self._device)[self.index]
        return super()._calc_name()


class MaxDeviceParameterBank(MaxDeviceParameterBankBase):
    def _collect_parameters(self):
        bank_count = self._banking_info.device_bank_count(self._device)
        if bank_count == 0:
            return [(None, None)] * self._size
        bank = self._get_parameters_for_bank_index(self.index)
        next_index = self.index + 1
        if self._size > DEFAULT_BANK_SIZE:
            if next_index < bank_count:
                bank.extend(self._get_parameters_for_bank_index(next_index))
        return bank

    def _get_parameters_for_bank_index(self, bank_index):
        parameters = self._device.parameters
        mx_index = bank_index - int(self._banking_info.has_main_bank(self._device))
        indices = self.device.get_bank_parameters(mx_index)
        parameters = [parameters[index] if index >= 0 else None for index in indices]
        return [(param, None) for param in parameters]

    def _calc_name(self):
        bank_count = self._banking_info.device_bank_count(self._device)
        if self._size > DEFAULT_BANK_SIZE:
            if self.index < bank_count:
                return self._banking_info.device_bank_names(self._device)[self.index]
        return super()._calc_name()


class BankingInfo(BankingInfoBase):
    def __init__(self, bank_definitions, bank_size=DEFAULT_BANK_SIZE, *a, **k):
        (super().__init__)(bank_definitions, *a, **k)
        self._bank_size = bank_size
        self._num_simultaneous_banks = 2 if bank_size > DEFAULT_BANK_SIZE else 1

    @property
    def bank_size(self):
        return self._bank_size

    @property
    def num_simultaneous_banks(self):
        return self._num_simultaneous_banks

    def device_bank_count(self, device, **k):
        if self._bank_size < DEFAULT_BANK_SIZE:
            return ceil(float(len(all_parameters(device))) / self._bank_size)
        return (super().device_bank_count)(device, **k)

    def device_bank_names(self, device, **k):
        if self._bank_size < DEFAULT_BANK_SIZE:
            return [BANK_FORMAT % (index + 1) for index in range(self.device_bank_count(device))]
        names = (super().device_bank_names)(device, **k)
        if self._num_simultaneous_banks == 2:
            if len(names) > 1:
                result = [BANK_NAME_JOIN_STR.join(n) for n in [(names[i], names[i + 1]) for i in range(len(names) - 1)]]
                result.append(names[-1])
                return result
        return names



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/control/toggle_button.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 2973 bytes
from .button import ButtonControlBase
from .control import Connectable, control_color, control_event


class ToggleButtonControl(ButtonControlBase):
    toggled = control_event("toggled")

    class State(ButtonControlBase.State, Connectable):
        untoggled_color = control_color("DefaultButton.Off")
        toggled_color = control_color("DefaultButton.On")
        requires_listenable_connected_property = True

        def __init__(self, untoggled_color=None, toggled_color=None, *a, **k):
            (super(ToggleButtonControl.State, self).__init__)(*a, **k)
            if untoggled_color is not None:
                self.untoggled_color = untoggled_color
            if toggled_color is not None:
                self.toggled_color = toggled_color
            self._is_toggled = False

        @property
        def is_toggled(self):
            return self._is_toggled

        @is_toggled.setter
        def is_toggled(self, toggled):
            if self._is_toggled != toggled:
                self._is_toggled = toggled
                self._send_current_color()

        def connect_property(self, *a):
            (super(ToggleButtonControl.State, self).connect_property)(*a)
            self.is_toggled = self.connected_property_value

        def on_connected_property_changed(self, value):
            self.is_toggled = value

        def _send_button_color(self):
            self._control_element.set_light(self.toggled_color if self._is_toggled else self.untoggled_color)

        def _on_pressed(self):
            super(ToggleButtonControl.State, self)._on_pressed()
            self._is_toggled = not self._is_toggled
            self._call_listener("toggled", self._is_toggled)
            self.connected_property_value = self.is_toggled



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/device_decorators.py
# Compiled at: 2023-12-21 15:35:35
# Size of source mod 2**32: 3205 bytes
from ableton.v2.control_surface import DelayDeviceDecorator, LiveObjectDecorator, SimplerDeviceDecorator
from ableton.v2.control_surface import DeviceDecoratorFactory as DeviceDecoratorFactoryBase


class DeviceDecorator(LiveObjectDecorator):
    def __init__(self, *a, **k):
        (super().__init__)(*a, **k)
        self.create_additional_parameters()
        self.register_disconnectables(self.options)

    def create_additional_parameters(self):
        raise NotImplementedError("Must implement this method to create additional parameters for the device")


class TransmuteDeviceDecorator(DeviceDecorator):
    def create_additional_parameters(self):
        self._add_non_automatable_enum_parameter(
            name="Hz/Note Mode",
            list="frequency_dial_mode_list",
            index="frequency_dial_mode",
        )
        self._add_non_automatable_enum_parameter(name="Pitch Mode", list="pitch_mode_list", index="pitch_mode")


class DriftDeviceDecorator(DeviceDecorator):
    def create_additional_parameters(self):
        self._add_non_automatable_enum_parameter(name="Voice Mode", list="voice_mode_list", index="voice_mode_index")
        self._add_non_automatable_enum_parameter(name="Voice Count", list="voice_count_list", index="voice_count_index")
        self._add_non_automatable_enum_parameter(
            name="LP Mod Src 1",
            list="mod_matrix_filter_source_1_list",
            index="mod_matrix_filter_source_1_index",
        )
        self._add_non_automatable_enum_parameter(
            name="LP Mod Src 2",
            list="mod_matrix_filter_source_2_list",
            index="mod_matrix_filter_source_2_index",
        )


class RoarDeviceDecorator(DeviceDecorator):
    def create_additional_parameters(self):
        self._add_non_automatable_enum_parameter(name="Routing", list="routing_mode_list", index="routing_mode_index")


class DeviceDecoratorFactory(DeviceDecoratorFactoryBase):
    DECORATOR_CLASSES = {
        "Delay": DelayDeviceDecorator,
        "Drift": DriftDeviceDecorator,
        "OriginalSimpler": SimplerDeviceDecorator,
        "Roar": RoarDeviceDecorator,
        "Transmute": TransmuteDeviceDecorator,
    }



from __future__ import annotations

from typing import Any, Callable

from Live import *
from Live.Application import *
from Live.Base import *
from Live.Browser import *
from Live.CcControlDevice import *
from Live.Chain import *
from Live.ChainMixerDevice import *
from Live.Clip import *
from Live.ClipSlot import *
from Live.CompressorDevice import *
from Live.Conversions import *
from Live.Device import *
from Live.DeviceIO import *
from Live.DeviceParameter import *
from Live.DriftDevice import *
from Live.DrumChain import *
from Live.DrumPad import *
from Live.Eq8Device import *
from Live.Groove import *
from Live.GroovePool import *
from Live.HybridReverbDevice import *
from Live.LomObject import *
from Live.MaxDevice import *
from Live.MeldDevice import *
from Live.MidiMap import *
from Live.MixerDevice import *
from Live.PluginDevice import *
from Live.RackDevice import *
from Live.RoarDevice import *
from Live.Sample import *
from Live.Scene import *
from Live.ShifterDevice import *
from Live.SimplerDevice import *
from Live.Song import *
from Live.SpectralResonatorDevice import *
from Live.Track import *
from Live.WavetableDevice import *


class ListenerHandle:
    """
    This class represents a Python listener when connected to a Live property.
    """

    def disconnect(self) -> None:
        """
        disconnect( (ListenerHandle)arg1) -> None :
            Disconnects the listener from its property

            C++ signature :
                void disconnect(LPythonRemote {lvalue})
        """

    @property
    def listener_func(self) -> Any:
        """
        Returns the original function
        """

    @property
    def listener_self(self) -> Any:
        """
        Returns the weak reference to original self, if it was a bound method
        """

    @property
    def name(self) -> str:
        """
        Prints the name of the property that this listener is connected to
        """


class ListenerVector:
    """
    A read only container for accessing a list of listeners.
    """

    def append(self, arg2: object) -> None:
        """
        append( (ListenerVector)arg1, (object)arg2) -> None :

            C++ signature :
                void append(std::__1::vector<TWeakPtr<LPythonRemote>, std::__1::allocator<TWeakPtr<LPythonRemote>>> {lvalue},boost::python::api::object)
        """

    def extend(self, arg2: object) -> None:
        """
        extend( (ListenerVector)arg1, (object)arg2) -> None :

            C++ signature :
                void extend(std::__1::vector<TWeakPtr<LPythonRemote>, std::__1::allocator<TWeakPtr<LPythonRemote>>> {lvalue},boost::python::api::object)
        """



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/elements/color.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 3823 bytes
from abc import ABC, abstractmethod
from typing import NamedTuple, Optional

from ...base import memoize, old_hasattr


@memoize
def create_rgb_color(values):
    if values is not None:
        return RgbColor(*values)
    return None


class Color(ABC):
    @abstractmethod
    def draw(self, interface):
        pass

    @property
    def midi_value(self):
        raise NotImplementedError


class SimpleColor(Color):
    def __init__(self, value, channel=None, *a, **k):
        (super().__init__)(*a, **k)
        self._value = value
        self._channel = channel

    @property
    def midi_value(self):
        return self._value

    def draw(self, interface):
        interface.send_value((self._value), channel=(self._channel))


class RgbColor(Color):
    def __init__(self, *values, **k):
        (super().__init__)(**k)
        self._values = values

    def draw(self, interface):
        (interface.send_value)(*self._values)


class ColorPart(NamedTuple):
    value: int
    channel = None
    channel: Optional[int]


class ComplexColor(Color):
    def __init__(self, color_parts, *a, **k):
        (super().__init__)(*a, **k)
        self._color_parts = color_parts

    def draw(self, interface):
        for part in self._color_parts:
            interface.send_value((part.value), channel=(part.channel))


class FallbackColor(Color):
    def __init__(self, rgb_color, fallback_color):
        self._rgb_color = rgb_color
        self._fallback_color = fallback_color

    @property
    def midi_value(self):
        return self._rgb_color.midi_value

    def draw(self, interface):
        if old_hasattr(interface, "is_rgb") and interface.is_rgb:
            self._rgb_color.draw(interface)
        else:
            self._fallback_color.draw(interface)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/_Framework/SysexValueControl.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 1115 bytes
from .InputControlElement import MIDI_SYSEX_TYPE, InputControlElement


class SysexValueControl(InputControlElement):
    def __init__(self, message_prefix=None, value_enquiry=None, default_value=None, *a, **k):
        (super().__init__)(a, msg_type=MIDI_SYSEX_TYPE, sysex_identifier=message_prefix, **k)
        self._value_enquiry = value_enquiry
        self._default_value = default_value

    def send_value(self, value_bytes):
        self.send_midi(self.message_sysex_identifier() + value_bytes + (247,))

    def enquire_value(self):
        self.send_midi(self._value_enquiry)

    def reset(self):
        if self._default_value is not None:
            self.send_value(self._default_value)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/layer.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 695 bytes
from ableton.v2.control_surface import Layer as LayerBaseClass


class Layer(LayerBaseClass):
    def on_received(self, client, *a, **k):
        (super().on_received)(client, *a, **k)
        client.num_layers += 1

    def on_lost(self, client):
        super().on_lost(client)
        client.num_layers -= 1



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/components/grid_resolution.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 3997 bytes
from typing import NamedTuple

from Live.Clip import GridQuantization

from ...base import listenable_property
from .. import Component
from ..controls import FixedRadioButtonGroup


class GridResolution(NamedTuple):
    name: str
    step_length: float
    grid: int
    is_triplet: bool


GRID_RESOLUTIONS = (
    GridResolution("1/32t", 0.08333333333333333, GridQuantization.g_thirtysecond, True),
    GridResolution("1/32", 0.125, GridQuantization.g_thirtysecond, False),
    GridResolution("1/16t", 0.16666666666666666, GridQuantization.g_sixteenth, True),
    GridResolution("1/16", 0.25, GridQuantization.g_sixteenth, False),
    GridResolution("1/8t", 0.3333333333333333, GridQuantization.g_eighth, True),
    GridResolution("1/8", 0.5, GridQuantization.g_eighth, False),
    GridResolution("1/4t", 0.6666666666666666, GridQuantization.g_quarter, True),
    GridResolution("1/4", 1.0, GridQuantization.g_quarter, False),
)
DEFAULT_INDEX = 3


class GridResolutionComponent(Component):
    resolution_buttons = FixedRadioButtonGroup(
        checked_color="NoteEditor.Resolution.Selected",
        unchecked_color="NoteEditor.Resolution.NotSelected",
        control_count=8,
    )

    def __init__(self, name="Grid_Resolution", resolutions=None, default_index=DEFAULT_INDEX, *a, **k):
        (super().__init__)(a, name=name, **k)
        self._resolutions = resolutions or GRID_RESOLUTIONS
        self._index = default_index
        self._update_resolution_buttons()

    @property
    def step_length(self):
        return self._resolutions[self.index].step_length

    @property
    def clip_grid(self):
        resolution = self._resolutions[self.index]
        return (resolution.grid, resolution.is_triplet)

    @property
    def is_triplet(self):
        return self._resolutions[self.index].is_triplet

    @listenable_property
    def index(self):
        return self._index

    @index.setter
    def index(self, index):
        if index != self._index:
            self._index = index
            self._update_resolution_buttons()
            self.notify_index()

    def set_to(self, name):
        name = name.lower()
        for index, grid_resolution in enumerate(self._resolutions):
            if grid_resolution.name == name:
                self.index = index
                return

        raise AssertionError(f"{name} is not a valid GridResolution name")

    @resolution_buttons.checked
    def resolution_buttons(self, button):
        self.index = button.index

    def _update_resolution_buttons(self):
        self.resolution_buttons.checked_index = self._index



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/_Framework/TrackArmState.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 1479 bytes
from .SubjectSlot import SlotManager, Subject, subject_slot


class TrackArmState(Subject, SlotManager):
    __subject_events__ = ("arm",)

    def __init__(self, track=None, *a, **k):
        (super().__init__)(*a, **k)
        self.set_track(track)

    def set_track(self, track):
        self._track = track
        self._arm = track and track.can_be_armed and (track.arm or track.implicit_arm)
        subject = track if track and (track.can_be_armed) else None
        self._on_explicit_arm_changed.subject = subject
        self._on_implicit_arm_changed.subject = subject

    @subject_slot("arm")
    def _on_explicit_arm_changed(self):
        self._on_arm_changed()

    @subject_slot("implicit_arm")
    def _on_implicit_arm_changed(self):
        self._on_arm_changed()

    def _on_arm_changed(self):
        new_state = self._track.arm or self._track.implicit_arm
        if self._arm != new_state:
            self._arm = new_state
            self.notify_arm()

    def _get_arm(self):
        if self._track.can_be_armed:
            return self._arm
        return False

    def _set_arm(self, new_state):
        if self._track.can_be_armed:
            self._track.arm = new_state
            if not new_state:
                self._track.implicit_arm = False
        self._arm = new_state

    arm = property(_get_arm, _set_arm)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/message_scheduler.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 9549 bytes
from collections import namedtuple

from ableton.v2.base import const


class MessageScheduler:
    def __init__(self, send_message_callback, timer, on_state_changed_callback=const(None)):
        self._send_message_callback = send_message_callback
        self._on_state_changed_callback = on_state_changed_callback
        self._timer = timer
        self._state = "idle"
        self._owner = None
        self._request_type = namedtuple("Request", "action owner message timeout")
        self._request_queue = []

    @property
    def is_idling(self):
        return self._state == "idle" and self._owner is None and len(self._request_queue) == 0

    def __repr__(self):
        return f"MessageScheduler(state={self._state}, owner={self._owner})"

    def _set_state(self, new_state):
        if new_state != self._state:
            self._on_state_changed_callback(new_state)
        self._state = new_state

    def _process_request(self, request):
        if request.action == "send":
            if self._state != "idle":
                if self._state != "grabbed" or self._owner == request.owner:
                    self._send_message_callback(request.message)
                    return True
                return False
            return None
        else:
            if request.action == "grab":
                if self._state == "idle":
                    self._set_state("grabbed")
                    self._owner = request.owner
                    self._owner.send_reply("grab", "1")
                    return True
                if self._state == "grabbed":
                    request.owner.report_error("unexpected grab")
                    return True
                return False
            else:
                if request.action == "release":
                    if self._state == "idle":
                        request.owner.report_error("unexpected release")
                        return True
                    if self._state == "grabbed":
                        self._owner.send_reply("release", "1")
                        self._set_state("idle")
                        self._owner = None
                        return True
                    return False
                else:
                    if request.action == "send_receive":
                        if self._state == "idle":
                            self._send_message_callback(request.message)
                            self._set_state("wait")
                            self._owner = request.owner
                            self._timer.start(request.timeout, self.handle_timeout)
                            return True
                        if self._state == "grabbed":
                            if self._owner == request.owner:
                                self._send_message_callback(request.message)
                                self._set_state("grabbed_wait")
                                self._timer.start(request.timeout, self.handle_timeout)
                                return True
                        return False
                    return None

    def _queue(self, request):
        if request.owner is not None:
            self._request_queue.append(request)

    def _process_single_request(self):
        for i, request in enumerate(self._request_queue):
            if self._owner in (None, request.owner):
                if self._process_request(request):
                    del self._request_queue[i]
                    return True
                else:
                    return False

        return False

    def _process_queue(self):
        while self._process_single_request():
            pass

    def send(self, owner, message):
        request = self._request_type("send", owner, message, 0)
        self._queue(request)
        self._process_queue()

    def grab(self, owner):
        request = self._request_type("grab", owner, None, 0)
        self._queue(request)
        self._process_queue()

    def release(self, owner):
        request = self._request_type("release", owner, None, 0)
        self._queue(request)
        self._process_queue()

    def send_receive(self, owner, message, timeout):
        request = self._request_type("send_receive", owner, message, timeout)
        self._queue(request)
        self._process_queue()

    def handle_message(self, message):
        if self._state == "idle":
            pass
        else:
            if self._state == "wait":
                if self._owner.is_expected_reply(message):
                    self._owner.send_reply("send_receive", message)
                    self._set_state("idle")
                    self._owner = None
                    self._timer.cancel()
                    self._process_queue()
            else:
                if self._state == "grabbed":
                    self._owner.send_reply("received", message)
                else:
                    if self._state == "grabbed_wait":
                        if self._owner.is_expected_reply(message):
                            self._owner.send_reply("send_receive", message)
                            self._set_state("grabbed")
                            self._timer.cancel()
                            self._process_queue()
                        else:
                            self._owner.send_reply("received", message)

    def handle_timeout(self):
        if self._state == "wait":
            self._owner.send_reply("send_receive", "timeout")
            self._set_state("idle")
            self._owner = None
            self._process_queue()
        else:
            if self._state == "grabbed_wait":
                self._owner.send_reply("send_receive", "timeout")
                self._set_state("grabbed")
                self._process_queue()

    def disconnect(self, owner):
        if self._state != "idle":
            self._request_queue = [r for r in self._request_queue if r.owner != owner]
            if self._owner == owner:
                self._owner = None
                self._set_state("idle")
                self._timer.cancel()
            self._process_queue()



from __future__ import annotations

from typing import Any, Callable

from Live import *
from Live.Application import *
from Live.Base import *
from Live.Browser import *
from Live.CcControlDevice import *
from Live.Chain import *
from Live.ChainMixerDevice import *
from Live.Clip import *
from Live.ClipSlot import *
from Live.CompressorDevice import *
from Live.Conversions import *
from Live.Device import *
from Live.DeviceIO import *
from Live.DeviceParameter import *
from Live.DriftDevice import *
from Live.DrumChain import *
from Live.DrumPad import *
from Live.Eq8Device import *
from Live.GroovePool import *
from Live.HybridReverbDevice import *
from Live.Listener import *
from Live.LomObject import *
from Live.MaxDevice import *
from Live.MeldDevice import *
from Live.MidiMap import *
from Live.MixerDevice import *
from Live.PluginDevice import *
from Live.RackDevice import *
from Live.RoarDevice import *
from Live.Sample import *
from Live.Scene import *
from Live.ShifterDevice import *
from Live.SimplerDevice import *
from Live.Song import *
from Live.SpectralResonatorDevice import *
from Live.Track import *
from Live.WavetableDevice import *


class Base:
    """
    Class that represent an enumeration of values for Base
    int([x]) -> integer
    int(x, base=10) -> integer

    Convert a number or string to an integer, or return 0 if no arguments
    are given.  If x is a number, return x.__int__().  For floating point
    numbers, this truncates towards zero.

    If x is not a number or if base is given, then x must be a string,
    bytes, or bytearray instance representing an integer literal in the
    given base.  The literal can be preceded by '+' or '-' and be surrounded
    by whitespace.  The base defaults to 10.  Valid bases are 0 and 2-36.
    Base 0 means to interpret the base from the string as an integer literal.
    >>> int('0b100', base=0)
    4
    """

    gb_four = 0
    gb_eight = 1
    gb_eight_triplet = 2
    gb_sixteen = 3
    gb_sixteen_triplet = 4
    gb_thirtytwo = 5
    count = 6


class Groove:
    """
    This class represents a groove in Live.
    """

    def add_name_listener(self, listener: Callable) -> None:
        """
        add_name_listener( (Groove)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "name" has changed.

            C++ signature :
                void add_name_listener(TPyHandle<AGroove>,boost::python::api::object)
        """

    def add_quantization_amount_listener(self, listener: Callable) -> None:
        """
        add_quantization_amount_listener( (Groove)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "quantization_amount" has changed.

            C++ signature :
                void add_quantization_amount_listener(TPyHandle<AGroove>,boost::python::api::object)
        """

    def add_random_amount_listener(self, listener: Callable) -> None:
        """
        add_random_amount_listener( (Groove)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "random_amount" has changed.

            C++ signature :
                void add_random_amount_listener(TPyHandle<AGroove>,boost::python::api::object)
        """

    def add_timing_amount_listener(self, listener: Callable) -> None:
        """
        add_timing_amount_listener( (Groove)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "timing_amount" has changed.

            C++ signature :
                void add_timing_amount_listener(TPyHandle<AGroove>,boost::python::api::object)
        """

    def add_velocity_amount_listener(self, listener: Callable) -> None:
        """
        add_velocity_amount_listener( (Groove)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "velocity_amount" has changed.

            C++ signature :
                void add_velocity_amount_listener(TPyHandle<AGroove>,boost::python::api::object)
        """

    def name_has_listener(self, listener: Callable) -> bool:
        """
        name_has_listener( (Groove)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "name".

            C++ signature :
                bool name_has_listener(TPyHandle<AGroove>,boost::python::api::object)
        """

    def quantization_amount_has_listener(self, listener: Callable) -> bool:
        """
        quantization_amount_has_listener( (Groove)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "quantization_amount".

            C++ signature :
                bool quantization_amount_has_listener(TPyHandle<AGroove>,boost::python::api::object)
        """

    def random_amount_has_listener(self, listener: Callable) -> bool:
        """
        random_amount_has_listener( (Groove)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "random_amount".

            C++ signature :
                bool random_amount_has_listener(TPyHandle<AGroove>,boost::python::api::object)
        """

    def remove_name_listener(self, listener: Callable) -> None:
        """
        remove_name_listener( (Groove)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "name".

            C++ signature :
                void remove_name_listener(TPyHandle<AGroove>,boost::python::api::object)
        """

    def remove_quantization_amount_listener(self, listener: Callable) -> None:
        """
        remove_quantization_amount_listener( (Groove)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "quantization_amount".

            C++ signature :
                void remove_quantization_amount_listener(TPyHandle<AGroove>,boost::python::api::object)
        """

    def remove_random_amount_listener(self, listener: Callable) -> None:
        """
        remove_random_amount_listener( (Groove)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "random_amount".

            C++ signature :
                void remove_random_amount_listener(TPyHandle<AGroove>,boost::python::api::object)
        """

    def remove_timing_amount_listener(self, listener: Callable) -> None:
        """
        remove_timing_amount_listener( (Groove)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "timing_amount".

            C++ signature :
                void remove_timing_amount_listener(TPyHandle<AGroove>,boost::python::api::object)
        """

    def remove_velocity_amount_listener(self, listener: Callable) -> None:
        """
        remove_velocity_amount_listener( (Groove)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "velocity_amount".

            C++ signature :
                void remove_velocity_amount_listener(TPyHandle<AGroove>,boost::python::api::object)
        """

    def timing_amount_has_listener(self, listener: Callable) -> bool:
        """
        timing_amount_has_listener( (Groove)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "timing_amount".

            C++ signature :
                bool timing_amount_has_listener(TPyHandle<AGroove>,boost::python::api::object)
        """

    def velocity_amount_has_listener(self, listener: Callable) -> bool:
        """
        velocity_amount_has_listener( (Groove)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "velocity_amount".

            C++ signature :
                bool velocity_amount_has_listener(TPyHandle<AGroove>,boost::python::api::object)
        """

    @property
    def base(self) -> Any:
        """
        Get/set the groove's base grid.
        """

    @base.setter
    def base(self, value: Any) -> None:
        pass

    @property
    def canonical_parent(self) -> Any:
        """
        Get the canonical parent of the groove.
        """

    @property
    def name(self) -> str:
        """
        Read/write/listen access to the groove's name
        """

    @name.setter
    def name(self, value: str) -> None:
        pass

    @property
    def quantization_amount(self) -> Any:
        """
        Read/write/listen access to the groove's quantization amount.
        """

    @quantization_amount.setter
    def quantization_amount(self, value: Any) -> None:
        pass

    @property
    def random_amount(self) -> Any:
        """
        Read/write/listen access to the groove's random amount.
        """

    @random_amount.setter
    def random_amount(self, value: Any) -> None:
        pass

    @property
    def timing_amount(self) -> Any:
        """
        Read/write/listen access to the groove's timing amount.
        """

    @timing_amount.setter
    def timing_amount(self, value: Any) -> None:
        pass

    @property
    def velocity_amount(self) -> Any:
        """
        Read/write/listen access to the groove's velocity amount.
        """

    @velocity_amount.setter
    def velocity_amount(self, value: Any) -> None:
        pass



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/midi.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 1570 bytes
from numbers import Number

NOTE_ON_STATUS = 144
NOTE_OFF_STATUS = 128
CC_STATUS = 176
PB_STATUS = 224
SYSEX_START = 240
SYSEX_END = 247
SYSEX_GENERAL_INFO = 6
SYSEX_NON_REALTIME = 126
SYSEX_IDENTITY_REQUEST_ID = 1
SYSEX_IDENTITY_RESPONSE_ID = 2
SYSEX_IDENTITY_REQUEST_MESSAGE = (
    SYSEX_START,
    SYSEX_NON_REALTIME,
    127,
    SYSEX_GENERAL_INFO,
    SYSEX_IDENTITY_REQUEST_ID,
    SYSEX_END,
)


def is_valid_channel(channel):
    if not isinstance(channel, Number):
        return False
    return 0 <= channel < 16


def is_valid_value(value, max_val=128):
    if not isinstance(value, Number):
        return False
    return 0 <= value < max_val


def is_valid_identifier(identifier):
    if not isinstance(identifier, Number):
        return False
    return 0 <= identifier < 128


def is_sysex(midi_bytes):
    return len(midi_bytes) != 3


def is_valid_sysex(midi_bytes):
    return is_sysex(midi_bytes) and midi_bytes[0] == SYSEX_START and midi_bytes[-1] == SYSEX_END


def is_pitchbend(midi_bytes):
    return midi_bytes[0] & 240 == PB_STATUS


def extract_value(midi_bytes):
    if is_pitchbend(midi_bytes):
        return midi_bytes[1] + (midi_bytes[2] << 7)
    return midi_bytes[2]


def pretty_print_bytes(midi_bytes):
    hex_values = list(map(hex, midi_bytes))
    return " ".join(hex_values)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/device_parameter_bank.py
# Compiled at: 2023-12-21 15:35:35
# Size of source mod 2**32: 6834 bytes
from ..base import EventObject, clamp, find_if, listenable_property, listens, listens_group, liveobj_valid, old_hasattr
from .banking_util import BANK_FORMAT, BANK_MAIN_KEY, BANK_PARAMETERS_KEY, all_parameters


class DeviceParameterBank(EventObject):
    def __init__(self, size=None, device=None, banking_info=None, *a, **k):
        (super().__init__)(*a, **k)
        self._size = size
        self._device = device
        self._banking_info = banking_info
        self._index = 0
        self._parameters = None
        self._on_parameters_changed.subject = device
        self._update_parameters()

    def bank_count(self):
        return self._banking_info.device_bank_count((self._device), bank_size=(self._size))

    def _adjust_index(self, index):
        return clamp(index, 0, max(0, self.bank_count() - 1))

    def _is_index_valid(self, index):
        if index is None:
            return False
        return self.bank_count() > index

    def _get_index(self):
        return self._index

    def _set_index(self, index):
        if self._index != index:
            if self._is_index_valid(index):
                index = self._adjust_index(index)
                self._index = index
                self._update_parameters()

    index = property(_get_index, _set_index)

    @listens("parameters")
    def _on_parameters_changed(self):
        self._index = self._adjust_index(self._index)
        self._update_parameters()

    @listenable_property
    def parameters(self):
        return self._parameters

    def _calc_name(self):
        return BANK_FORMAT % (self.index + 1)

    @property
    def name(self):
        if liveobj_valid(self._device):
            return self._calc_name()
        return ""

    @property
    def device(self):
        return self._device

    def _collect_parameters(self):
        parameters = all_parameters(self._device)
        offset = self._index * self._size
        params = parameters[offset:]
        params.extend([None] * (self._size - len(params)))
        return [(param, None) for param in params]

    def _update_parameters(self):
        parameters = self._collect_parameters()[: self._size]
        if self._parameters != parameters:
            self._parameters = parameters
            self.notify_parameters()


class DescribedDeviceParameterBank(DeviceParameterBank):
    def __init__(self, device=None, banking_info=None, *a, **k):
        self._definition = banking_info.device_bank_definition(device)
        self._dynamic_slots = []
        (super().__init__)(a, device=device, banking_info=banking_info, **k)
        self._update_parameters()

    @listens_group("content")
    def _on_slot_content_changed(self, _slot):
        self._update_parameters()

    def _current_parameter_slots(self):
        return self._definition.value_by_index(self.index).get(BANK_PARAMETERS_KEY) or ()

    def _content_slots(self):
        return self._current_parameter_slots()

    def _setup_dynamic_slots(self):
        for slot in self._dynamic_slots:
            slot.set_parameter_host(None)
            self.unregister_disconnectable(slot)

        self._dynamic_slots = list(filter(lambda s: old_hasattr(s, "notify_content"), self._content_slots()))
        for slot in self._dynamic_slots:
            self.register_disconnectable(slot)
            slot.set_parameter_host(self.device)

        self._on_slot_content_changed.replace_subjects(self._dynamic_slots)

    def _calc_name(self):
        return self._definition.key_by_index(self.index)

    def _collect_parameters(self):
        parameters = self._device.parameters
        bank_slots = self._current_parameter_slots()

        def _collect_parameter(slot_definition):
            parameter = find_if(lambda p: p.original_name == str(slot_definition), parameters)
            display_name_transformer = getattr(slot_definition, "display_name_transformer", None)
            return (
                parameter
                and (parameter, display_name_transformer and display_name_transformer(parameter.name))
                or (None, None)
            )

        return [_collect_parameter(slot_definition) for slot_definition in bank_slots]

    def _update_parameters(self):
        self._setup_dynamic_slots()
        super()._update_parameters()
        self._on_parameter_name_changed.replace_subjects([parameter for parameter, name in self._parameters])

    @listens_group("name")
    def _on_parameter_name_changed(self, parameter):
        self._update_parameters()


class MaxDeviceParameterBank(DeviceParameterBank):
    def __init__(self, *a, **k):
        (super().__init__)(*a, **k)
        self._on_bank_parameters_changed.subject = self.device

    def _calc_name(self):
        if self.bank_count() == 0:
            return BANK_MAIN_KEY
        mx_index = self.index - int(self._banking_info.has_main_bank(self._device))
        provided_name = self.device.get_bank_name(mx_index)
        if len(provided_name) > 0:
            return provided_name
        return super()._calc_name()

    @listens("bank_parameters_changed")
    def _on_bank_parameters_changed(self):
        self._update_parameters()

    def _collect_parameters(self):
        if self.bank_count() == 0:
            return [(None, None)] * self._size
        parameters = self._device.parameters
        mx_index = self.index - int(self._banking_info.has_main_bank(self._device))
        indices = self.device.get_bank_parameters(mx_index)
        parameters = [parameters[index] if index >= 0 else None for index in indices]
        return [(param, None) for param in parameters]


def create_device_bank(device, banking_info):
    bank = None
    if liveobj_valid(device):
        if banking_info.has_bank_count(device):
            bank_class = MaxDeviceParameterBank
        else:
            if banking_info.device_bank_definition(device) is not None:
                bank_class = DescribedDeviceParameterBank
            else:
                bank_class = DeviceParameterBank
        bank = bank_class(device=device, size=8, banking_info=banking_info)
    return bank
